"""
Platydock

Platydock is a helper that allows you to easily manage (create, destroy, start,
stop and list) Platypush instances as Docker images.

.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
.. license: MIT
"""

import argparse
import enum
import os
import pathlib
import re
import shutil
import subprocess
import sys
import textwrap
import traceback as tb
import yaml

from platypush.config import Config
from platypush.utils import manifest

workdir = os.path.join(os.path.expanduser('~'), '.local', 'share',
                       'platypush', 'platydock')


class Action(enum.Enum):
    build = 'build'
    start = 'start'
    stop = 'stop'
    rm = 'rm'
    ls = 'ls'

    def __str__(self):
        return self.value


def _parse_deps(cls):
    deps = []

    for line in cls.__doc__.split('\n'):
        m = re.search(r'\(``pip install (.+)``\)', line)
        if m:
            deps.append(m.group(1))

    return deps


def generate_dockerfile(deps, ports, cfgfile, device_dir, python_version):
    device_id = Config.get('device_id')
    if not device_id:
        raise RuntimeError(('You need to specify a device_id in {} - Docker ' +
                            'containers cannot rely on hostname').format(cfgfile))

    os.makedirs(device_dir, exist_ok=True)
    content = textwrap.dedent(
        '''
        FROM python:{python_version}-slim-bullseye

        RUN mkdir -p /app
        RUN mkdir -p /etc/platypush
        RUN mkdir -p /usr/local/share/platypush\n
        '''.format(python_version=python_version)).lstrip()

    srcdir = os.path.dirname(cfgfile)
    cfgfile_copy = os.path.join(device_dir, 'config.yaml')
    shutil.copy(cfgfile, cfgfile_copy, follow_symlinks=True)
    content += 'COPY config.yaml /etc/platypush/\n'
    backend_config = Config.get_backends()

    # Redis configuration for Docker
    if 'redis' not in backend_config:
        backend_config['redis'] = {
            'redis_args': {
                'host': 'redis',
                'port': 6379,
            }
        }

        with open(cfgfile_copy, 'a') as f:
            f.write('\n# Automatically added by platydock, do not remove\n' + yaml.dump({
                'backend.redis': backend_config['redis'],
            }) + '\n')

    # Main database configuration
    has_main_db = False
    with open(cfgfile_copy, 'r') as f:
        for line in f.readlines():
            if re.match(r'^(main.)?db.*', line):
                has_main_db = True
                break

    if not has_main_db:
        with open(cfgfile_copy, 'a') as f:
            f.write('\n# Automatically added by platydock, do not remove\n' + yaml.dump({
                'main.db': {
                    'engine': 'sqlite:////platypush.db',
                }
            }) + '\n')

    # Copy included files
    # noinspection PyProtectedMember
    for include in Config._included_files:
        incdir = os.path.relpath(os.path.dirname(include), srcdir)
        destdir = os.path.join(device_dir, incdir)
        pathlib.Path(destdir).mkdir(parents=True, exist_ok=True)
        shutil.copy(include, destdir, follow_symlinks=True)
        content += 'RUN mkdir -p /etc/platypush/' + incdir + '\n'
        content += 'COPY ' + os.path.relpath(include, srcdir) + \
                   ' /etc/platypush/' + incdir + '\n'

    # Copy script files
    scripts_dir = os.path.join(os.path.dirname(cfgfile), 'scripts')
    if os.path.isdir(scripts_dir):
        local_scripts_dir = os.path.join(device_dir, 'scripts')
        remote_scripts_dir = '/etc/platypush/scripts'
        shutil.copytree(scripts_dir, local_scripts_dir, symlinks=True, dirs_exist_ok=True)
        content += f'RUN mkdir -p {remote_scripts_dir}\n'
        content += f'COPY scripts/ {remote_scripts_dir}\n'

    packages = deps.pop('packages', None)
    pip = deps.pop('pip', None)
    exec_cmds = deps.pop('exec', None)
    pkg_cmd = f'\n\t&& apt-get install --no-install-recommends -y {" ".join(packages)} \\' if packages else ''
    pip_cmd = f'\n\t&& pip install {" ".join(pip)} \\' if pip else ''
    content += f'''
RUN dpkg --configure -a \\
    && apt-get -f install \\
    && apt-get --fix-missing install \\
    && apt-get clean \\
    && apt-get update \\
    && apt-get -y upgrade \\
    && apt-get -y dist-upgrade \\
    && apt-get install --no-install-recommends -y apt-utils \\
    && apt-get install --no-install-recommends -y build-essential \\
    && apt-get install --no-install-recommends -y git \\
    && apt-get install --no-install-recommends -y sudo \\
    && apt-get install --no-install-recommends -y libffi-dev \\
    && apt-get install --no-install-recommends -y libcap-dev \\
    && apt-get install --no-install-recommends -y libjpeg-dev \\{pkg_cmd}{pip_cmd}'''

    for exec_cmd in exec_cmds:
        content += f'\n\t&& {exec_cmd} \\'
    content += '''
    && apt-get install --no-install-recommends -y zlib1g-dev

RUN git clone --recursive https://git.platypush.tech/platypush/platypush.git /app \\
    && cd /app \\
    && pip install -r requirements.txt

RUN apt-get remove -y git \\
    && apt-get remove -y build-essential \\
    && apt-get remove -y libffi-dev \\
    && apt-get remove -y libjpeg-dev \\
    && apt-get remove -y libcap-dev \\
    && apt-get remove -y zlib1g-dev \\
    && apt-get remove -y apt-utils \\
    && apt-get clean \\
    && apt-get autoremove -y \\
    && rm -rf /var/lib/apt/lists/*
'''

    for port in ports:
        content += 'EXPOSE {}\n'.format(port)

    content += textwrap.dedent(
        '''

        ENV PYTHONPATH /app:$PYTHONPATH
        CMD ["python", "-m", "platypush"]
        ''')

    dockerfile = os.path.join(device_dir, 'Dockerfile')
    print('Generating Dockerfile {}'.format(dockerfile))

    with open(dockerfile, 'w') as f:
        f.write(content)


def build(args):
    global workdir

    ports = set()
    parser = argparse.ArgumentParser(
        prog='platydock build',
        description='Build a Platypush image from a config.yaml'
    )

    parser.add_argument('-c', '--config', type=str, required=True,
                        help='Path to the platypush configuration file')
    parser.add_argument('-p', '--python-version', type=str, default='3.9',
                        help='Python version to be used')

    opts, args = parser.parse_known_args(args)

    cfgfile = os.path.abspath(os.path.expanduser(opts.config))
    manifest._available_package_manager = 'apt'  # Force apt for Debian-based Docker images
    install_cmds = manifest.get_dependencies_from_conf(cfgfile)
    python_version = opts.python_version
    backend_config = Config.get_backends()

    # Container exposed ports
    if backend_config.get('http'):
        from platypush.backend.http import HttpBackend
        # noinspection PyProtectedMember
        ports.add(backend_config['http'].get('port', HttpBackend._DEFAULT_HTTP_PORT))
        # noinspection PyProtectedMember
        ports.add(backend_config['http'].get('websocket_port', HttpBackend._DEFAULT_WEBSOCKET_PORT))

    if backend_config.get('tcp'):
        ports.add(backend_config['tcp']['port'])

    if backend_config.get('websocket'):
        from platypush.backend.websocket import WebsocketBackend
        # noinspection PyProtectedMember
        ports.add(backend_config['websocket'].get('port', WebsocketBackend._default_websocket_port))

    dev_dir = os.path.join(workdir, Config.get('device_id'))
    generate_dockerfile(
        deps=dict(install_cmds), ports=ports, cfgfile=cfgfile, device_dir=dev_dir, python_version=python_version
    )

    subprocess.call(
        ['docker', 'build', '-t', 'platypush-{}'.format(Config.get('device_id')), dev_dir]
    )


def start(args):
    global workdir

    parser = argparse.ArgumentParser(prog='platydock start',
                                     description='Start a Platypush container',
                                     epilog=textwrap.dedent('''
                                       You can append additional options that
                                       will be passed to the docker container.
                                       Example:

                                            --add-host='myhost:192.168.1.1'
                                       '''))

    parser.add_argument('image', type=str, help='Platypush image to start')
    parser.add_argument('-p', '--publish', action='append', nargs='*', default=[],
                        help=textwrap.dedent('''
                                             Container's ports to expose to the host.
                                             Note that the default exposed ports from
                                             the container service will be exposed unless
                                             these mappings override them (e.g. port 8008
                                             on the container will be mapped to 8008 on
                                             the host).

                                             Example:

                                             -p 18008:8008 -p 18009:8009
                                             '''))

    parser.add_argument('-a', '--attach', action='store_true', default=False,
                        help=textwrap.dedent('''
                                             If set, then attach to the container after starting it up (default: false).
                                             '''))

    opts, args = parser.parse_known_args(args)
    ports = {}
    dockerfile = os.path.join(workdir, opts.image, 'Dockerfile')

    with open(dockerfile) as f:
        for line in f:
            m = re.match(r'expose (\d+)', line.strip().lower())
            if m:
                ports[m.group(1)] = m.group(1)

    for mapping in opts.publish:
        host_port, container_port = mapping[0].split(':')
        ports[container_port] = host_port

    print('Preparing Redis support container')
    subprocess.call(['docker', 'pull', 'redis'])
    subprocess.call(['docker', 'run', '--rm', '--name', 'redis-' + opts.image,
                     '-d', 'redis'])

    docker_cmd = ['docker', 'run', '--rm', '--name', opts.image, '-it',
                  '--link', 'redis-' + opts.image + ':redis']

    for container_port, host_port in ports.items():
        docker_cmd += ['-p', host_port + ':' + container_port]

    docker_cmd += args
    docker_cmd += ['-d', 'platypush-' + opts.image]

    print('Starting Platypush container {}'.format(opts.image))
    subprocess.call(docker_cmd)

    if opts.attach:
        subprocess.call(['docker', 'attach', opts.image])


def stop(args):
    parser = argparse.ArgumentParser(prog='platydock stop',
                                     description='Stop a Platypush container')

    parser.add_argument('container', type=str, help='Platypush container to stop')
    opts, args = parser.parse_known_args(args)

    print('Stopping Platypush container {}'.format(opts.container))
    subprocess.call(['docker', 'kill', '{}'.format(opts.container)])

    print('Stopping Redis support container')
    subprocess.call(['docker', 'stop', 'redis-{}'.format(opts.container)])


def rm(args):
    global workdir

    parser = argparse.ArgumentParser(prog='platydock rm',
                                     description='Remove a Platypush image. ' +
                                                 'NOTE: make sure that no container is ' +
                                                 'running nor linked to the image before ' +
                                                 'removing it')

    parser.add_argument('image', type=str, help='Platypush image to remove')
    opts, args = parser.parse_known_args(args)

    subprocess.call(['docker', 'rmi', 'platypush-{}'.format(opts.image)])
    shutil.rmtree(os.path.join(workdir, opts.image), ignore_errors=True)


def ls(args):
    parser = argparse.ArgumentParser(prog='platydock ls',
                                     description='List available Platypush containers')
    parser.add_argument('filter', type=str, nargs='?',
                        help='Image name filter')

    opts, args = parser.parse_known_args(args)

    p = subprocess.Popen(['docker', 'images'], stdout=subprocess.PIPE)
    output = p.communicate()[0].decode().split('\n')
    header = output.pop(0)
    images = []

    for line in output:
        if re.match(r'^platypush-(.+?)\s.*', line):
            if not opts.filter or (opts.filter and opts.filter in line):
                images.append(line)

    if images:
        print(header)

        for image in images:
            print(image)


def main():
    parser = argparse.ArgumentParser(prog='platydock', add_help=False,
                                     description='Manage Platypush docker containers',
                                     epilog='Use platydock <action> --help to ' +
                                            'get additional help')

    # noinspection PyTypeChecker
    parser.add_argument('action', nargs='?', type=Action, choices=list(Action),
                        help='Action to execute')
    parser.add_argument('-h', '--help', action='store_true', help='Show usage')
    opts, args = parser.parse_known_args(sys.argv[1:])

    if (opts.help and not opts.action) or (not opts.help and not opts.action):
        parser.print_help()
        return 1

    globals()[str(opts.action)](sys.argv[2:])


if __name__ == '__main__':
    ERR_PREFIX = '\n\033[6;31;47mERROR\033[0;91m '
    ERR_SUFFIX = '\033[0m'

    try:
        main()
    except Exception as e:
        tb.print_exc(file=sys.stdout)
        print(ERR_PREFIX + str(e) + ERR_SUFFIX, file=sys.stderr)

# vim:sw=4:ts=4:et:



import asyncio
import logging
import threading
import time

from abc import ABC, abstractmethod
from functools import wraps
from typing import Optional

from platypush.bus import Bus
from platypush.common import ExtensionWithManifest
from platypush.event import EventGenerator
from platypush.message.response import Response
from platypush.utils import get_decorators, get_plugin_name_by_class, set_thread_name


def action(f):
    @wraps(f)
    def _execute_action(*args, **kwargs):
        response = Response()
        result = f(*args, **kwargs)

        if result and isinstance(result, Response):
            result.errors = (
                result.errors if isinstance(result.errors, list) else [result.errors]
            )
            response = result
        elif isinstance(result, tuple) and len(result) == 2:
            response.errors = result[1] if isinstance(result[1], list) else [result[1]]

            if len(response.errors) == 1 and response.errors[0] is None:
                response.errors = []
            response.output = result[0]
        else:
            response = Response(output=result, errors=[])

        return response

    # Propagate the docstring
    _execute_action.__doc__ = f.__doc__
    return _execute_action


class Plugin(EventGenerator, ExtensionWithManifest):  # lgtm [py/missing-call-to-init]
    """Base plugin class"""

    def __init__(self, **kwargs):
        super().__init__()
        self.logger = logging.getLogger(
            'platypush:plugin:' + get_plugin_name_by_class(self.__class__)
        )
        if 'logging' in kwargs:
            self.logger.setLevel(getattr(logging, kwargs['logging'].upper()))

        self.registered_actions = set(
            get_decorators(self.__class__, climb_class_hierarchy=True).get('action', [])
        )

    def run(self, method, *args, **kwargs):
        assert (
            method in self.registered_actions
        ), '{} is not a registered action on {}'.format(method, self.__class__.__name__)
        return getattr(self, method)(*args, **kwargs)


class RunnablePlugin(Plugin):
    """
    Class for runnable plugins - i.e. plugins that have a start/stop method and can be started.
    """

    def __init__(self, poll_interval: Optional[float] = None, **kwargs):
        """
        :param poll_interval: How often the :meth:`.loop` function should be execute (default: None, no pause/interval).
        """
        super().__init__(**kwargs)
        self.poll_interval = poll_interval
        self.bus: Optional[Bus] = None
        self._should_stop = threading.Event()
        self._thread: Optional[threading.Thread] = None

    def main(self):
        raise NotImplementedError()

    def should_stop(self):
        return self._should_stop.is_set()

    def wait_stop(self, timeout=None):
        return self._should_stop.wait(timeout=timeout)

    def start(self):
        set_thread_name(self.__class__.__name__)
        self._thread = threading.Thread(target=self._runner)
        self._thread.start()

    def stop(self):
        self._should_stop.set()
        if self._thread and self._thread.is_alive():
            self.logger.info(f'Waiting for {self.__class__.__name__} to stop')
            try:
                if self._thread:
                    self._thread.join()
            except Exception as e:
                self.logger.warning(f'Could not join thread on stop: {e}')

        self.logger.info(f'{self.__class__.__name__} stopped')

    def _runner(self):
        self.logger.info(f'Starting {self.__class__.__name__}')

        while not self.should_stop():
            try:
                self.main()
            except Exception as e:
                self.logger.exception(e)

            if self.poll_interval:
                time.sleep(self.poll_interval)

        self._thread = None


class AsyncRunnablePlugin(RunnablePlugin, ABC):
    """
    Class for runnable plugins with an asynchronous event loop attached.
    """

    def __init__(self, *args, _stop_timeout: Optional[float] = 30.0, **kwargs):
        super().__init__(*args, **kwargs)

        self._stop_timeout = _stop_timeout
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._loop_runner: Optional[threading.Thread] = None
        self._task: Optional[asyncio.Task] = None

    @property
    def _should_start_runner(self):
        return True

    @abstractmethod
    async def listen(self):
        pass

    async def _listen(self):
        try:
            await self.listen()
        except KeyboardInterrupt:
            pass
        except RuntimeError as e:
            if not (
                str(e).startswith('Event loop stopped before ')
                or str(e).startswith('no running event loop')
            ):
                raise e

    def _start_listener(self):
        set_thread_name(self.__class__.__name__ + ':listener')
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)

        self._task = self._loop.create_task(self._listen())
        if hasattr(self._task, 'set_name'):
            self._task.set_name(self.__class__.__name__ + '.listen')
        self._loop.run_forever()

    def main(self):
        if self.should_stop() or (self._loop_runner and self._loop_runner.is_alive()):
            self.logger.info('The main loop is already being run/stopped')
            return

        if self._should_start_runner:
            self._loop_runner = threading.Thread(target=self._start_listener)
            self._loop_runner.start()

        self.wait_stop()

    def stop(self):
        if self._task and self._loop and not self._task.done():
            self._loop.call_soon_threadsafe(self._task.cancel)

        if self._loop and self._loop.is_running():
            self._loop.call_soon_threadsafe(self._loop.stop)
            self._loop = None

        if self._loop_runner and self._loop_runner.is_alive():
            try:
                self._loop_runner.join(timeout=self._stop_timeout)
            finally:
                self._loop_runner = None

        super().stop()


# vim:sw=4:ts=4:et:



import base64
import datetime
import errno
import getpass
import os
import threading

from binascii import hexlify
from stat import S_ISDIR, S_ISREG, S_ISLNK, S_ISCHR, S_ISFIFO, S_ISSOCK, S_ISBLK, S_ISDOOR
from typing import Optional, Dict, Tuple, List, Union, Any

from paramiko import DSSKey, RSAKey, SSHClient, WarningPolicy, SFTPClient
from paramiko.py3compat import u

try:
    from paramiko import GSS_AUTH_AVAILABLE
except ImportError:
    from paramiko.ssh_gss import GSS_AUTH_AVAILABLE

from platypush import Response
from platypush.message.response.ssh import SSHKeygenResponse
from platypush.plugins import Plugin, action
from platypush.plugins.ssh.tunnel.forward import forward_tunnel
from platypush.plugins.ssh.tunnel.reverse import reverse_tunnel, close_tunnel


class SshPlugin(Plugin):
    """
    SSH plugin.

    Requires:

        * **paramiko** (``pip install paramiko``)

    """

    key_dispatch_table = {'dsa': DSSKey, 'rsa': RSAKey}

    def __init__(self, key_file: Optional[str] = None, passphrase: Optional[str] = None, **kwargs):
        """
        :param key_file: Default key file (default: any "id_rsa", "id_dsa", "id_ecdsa", or "id_ed25519" key discoverable
            in ``~/.ssh/``.
        :param passphrase: Key file passphrase (default: None).
        """
        super().__init__(**kwargs)
        self.key_file = os.path.abspath(os.path.expanduser(key_file)) if key_file else None
        self.passphrase = passphrase
        self._sessions: Dict[Tuple[str, int, Optional[str]], SSHClient] = {}
        self._fwd_tunnels: Dict[Tuple[int, str, int], dict] = {}
        self._rev_tunnels: Dict[Tuple[int, str, int], dict] = {}

    def _get_key(self, key_file: Optional[str] = None, passphrase: Optional[str] = None):
        key_file = key_file or self.key_file
        return (os.path.abspath(os.path.expanduser(key_file)) if key_file else None,
                passphrase or self.passphrase)

    @staticmethod
    def _get_host_port_user(host: str, port: int = 22, user: Optional[str] = None, **_):
        if host.find('@') >= 0:
            user, host = host.split('@')
        if host.find(':') >= 0:
            host, port = host.split(':')
        if not user:
            user = getpass.getuser()

        return host, int(port), user

    # noinspection PyShadowingBuiltins
    @action
    def keygen(self,
               filename: str,
               type: str = 'rsa',
               bits: int = 4096,
               comment: Optional[str] = None,
               passphrase: Optional[str] = None) -> SSHKeygenResponse:
        """
        Generate an SSH keypair.

        :param filename: Output file name for the private key (the public key will be stored in <filename>.pub).
        :param type: Encryption algorithm, either "rsa" or "dsa" (default: "rsa").
        :param bits: Key length in bits (default: 4096).
        :param comment: Key comment (default: None).
        :param passphrase: Key passphrase (default: None).
        :return: :class:`platypush.message.response.ssh.SSHKeygenResponse`.
        """
        assert type != 'dsa' or bits <= 1024, 'DSA keys support a maximum of 1024 bits'
        assert type in self.key_dispatch_table, 'No such type: {}. Available types: {}'.format(
            type, self.key_dispatch_table.keys())

        if filename:
            filename = os.path.abspath(os.path.expanduser(filename))

        prv = self.key_dispatch_table[type].generate(bits=bits)
        prv.write_private_key_file(filename=filename, password=passphrase)
        pub = self.key_dispatch_table[type](filename=filename, password=passphrase)
        pub_file = '{filename}.pub'.format(filename=filename)
        with open(pub_file, 'w') as f:
            f.write('{name} {key}'.format(name=pub.get_name(), key=pub.get_base64()))
            if comment:
                f.write(' ' + comment)

        hash = u(hexlify(pub.get_fingerprint()))
        return SSHKeygenResponse(fingerprint=hash, key_file=filename, pub_key_file=pub_file)

    def run(self, *args, **kwargs):
        try:
            return super().run(*args, **kwargs)
        except Exception as e:
            raise AssertionError(e)

    def _connect(self,
                 host: str,
                 port: int = 22,
                 user: Optional[str] = None,
                 password: Optional[str] = None,
                 key_file: Optional[str] = None,
                 passphrase: Optional[str] = None,
                 compress: bool = False,
                 timeout: Optional[int] = None,
                 auth_timeout: Optional[int] = None) -> SSHClient:
        try:
            host, port, user = self._get_host_port_user(host, port, user)
            key = (host, port, user)
            if key in self._sessions:
                self.logger.info('[Connect] The SSH session is already active: {user}@{host}:{port}'.format(
                    user=user, host=host, port=port))
                return self._sessions[key]

            key_file, passphrase = self._get_key(key_file, passphrase)
            client = SSHClient()
            client.set_missing_host_key_policy(WarningPolicy())

            args = {
                'hostname': host,
                'port': port,
                'username': user,
                'compress': compress,
                'timeout': timeout,
                'auth_timeout': auth_timeout,
            }

            if password:
                args['password'] = password
            elif key_file:
                args['key_filename'] = key_file
                args['passphrase'] = passphrase
                args['gss_auth'] = GSS_AUTH_AVAILABLE
                args['gss_kex'] = GSS_AUTH_AVAILABLE
            else:
                client.load_system_host_keys()

            client.connect(**args)
            self._sessions[key] = client
            return client
        except Exception as e:
            self.logger.exception(e)
            raise AssertionError('Connection to {} failed: {}'.format(host, str(e)))

    @action
    def connect(self,
                host: str,
                port: int = 22,
                user: Optional[str] = None,
                password: Optional[str] = None,
                key_file: Optional[str] = None,
                passphrase: Optional[str] = None,
                compress: bool = False,
                timeout: Optional[int] = None,
                auth_timeout: Optional[int] = None) -> None:
        """
        Open an SSH connection.

        :param host: Host name or IP. Can also be in the format ``[user]@<host>:[port]``.
        :param port: Remote port (default: 22).
        :param user: Username (default: None, same user name as the one running platypush).
        :param password: Password (default: None).
        :param key_file: Key file to use for authentication (default: None).
        :param passphrase: Passphrase for the key file (default: None).
        :param compress: Compress data on the connection (default: False).
        :param timeout: Data transfer timeout in seconds (default: None).
        :param auth_timeout: Authentication timeout in seconds (default: None).
        """
        self._connect(host=host, port=port, user=user, password=password, key_file=key_file, passphrase=passphrase,
                      compress=compress, timeout=timeout, auth_timeout=auth_timeout)

    @action
    def disconnect(self,
                   host: str,
                   port: int = 22,
                   user: Optional[str] = None) -> None:
        """
        Close a connection to a host.

        :param host: Host name or IP. Can also be in the format ``[user]@<host>:[port]``.
        :param port: Remote port (default: 22).
        :param user: Username (default: None, same user name as the one running platypush).
        """
        host, port, user = self._get_host_port_user(host, port, user)
        key = (host, port, user)
        if key not in self._sessions:
            self.logger.info('[Disconnect] The SSH session is not active: {user}@{host}:{port}'.format(
                user=user, host=host, port=port))

        session = self._sessions[key]
        try:
            session.close()
        except Exception as e:
            self.logger.exception(e)

        del self._sessions[key]

    @action
    def exec(self, cmd: str, keep_alive: bool = False, timeout: Optional[int] = None,
             stdin: Optional[str] = None, env: Optional[Dict[str, str]] = None, **kwargs) -> Response:
        """
        Run a command on a host.

        :param cmd: Command to run
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param timeout: Communication timeout in seconds (default: None).
        :param stdin: Optional string to pass on the stdin of the command.
        :param env: Dictionary of environment variables to be used for the connection (default: None).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        :return: The output of the executed command.
        """
        client = self._connect(**kwargs)

        def decode(buf: bytes) -> str:
            try:
                buf = buf.decode()
            except (ValueError, TypeError):
                buf = base64.encodebytes(buf).decode()

            if buf.endswith('\n'):
                buf = buf[:-1]
            return buf

        try:
            _in, _out, _err = client.exec_command(cmd, timeout=timeout, environment=env)
            if stdin:
                with _in:
                    _in.write(stdin)

            resp = Response()
            with _out:
                resp.output = decode(_out.read())
            with _err:
                err = decode(_err.read())
                if err:
                    resp.errors = [err]

            return resp
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @staticmethod
    def is_directory(sftp: SFTPClient, path: str) -> bool:
        f = sftp.lstat(path)
        return S_ISDIR(f.st_mode)

    @classmethod
    def sftp_walk(cls, sftp: SFTPClient, path: str):
        files = []
        folders = []

        for f in sftp.listdir_attr(path):
            if S_ISDIR(f.st_mode):
                folders.append(f.filename)
            else:
                files.append(f.filename)

        yield path, folders, files

        for folder in folders:
            new_path = os.path.join(path, folder)
            for x in cls.sftp_walk(sftp, new_path):
                yield x

    def sftp_get(self, sftp: SFTPClient, remote_path: str, local_path: str, recursive: bool = False) -> None:
        if self.is_directory(sftp, remote_path):
            assert recursive, '{} is a directory on the server but recursive has been set to False'
            local_path = os.path.join(local_path, os.path.basename(remote_path))
            os.makedirs(local_path, mode=0o755, exist_ok=True)
            sftp.chdir(remote_path)

            for path, folders, files in self.sftp_walk(sftp, '.'):
                new_local_path = os.path.join(local_path, path)
                os.makedirs(new_local_path, mode=0o755, exist_ok=True)

                for file in files:
                    self.logger.info('Downloading file {} from {} to {}'.format(file, path, new_local_path))
                    self.sftp_get(sftp,
                                  os.path.join(remote_path, path, file),
                                  os.path.join(new_local_path, file),
                                  recursive=recursive)
        else:
            if os.path.isdir(local_path):
                local_path = os.path.join(local_path, os.path.basename(remote_path))

            sftp.get(remote_path, local_path)

    @action
    def get(self, remote_path: str, local_path: str, recursive: bool = False, keep_alive: bool = False,
            **kwargs) -> None:
        """
        Download a file or folder from an SSH server.

        :param remote_path: Remote path (file or directory).
        :param local_path: Local path (file or directory).
        :param recursive: Set to True if you want to recursively download folders (default: False).
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        local_path = os.path.abspath(os.path.expanduser(local_path))
        kwargs['compress'] = True
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            self.sftp_get(sftp, remote_path=remote_path, local_path=local_path, recursive=recursive)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def put(self, remote_path: str, local_path: str, recursive: bool = False, keep_alive: bool = False,
            **kwargs) -> None:
        """
        Upload a file or folder to an SSH server.

        :param remote_path: Remote path (file or directory).
        :param local_path: Local path (file or directory).
        :param recursive: Set to True if you want to recursively upload folders (default: False).
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        local_path = os.path.abspath(os.path.expanduser(local_path))
        assert os.path.exists(local_path), os.strerror(errno.ENOENT)
        kwargs['compress'] = True
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            if os.path.isdir(local_path):
                try:
                    sftp.mkdir(remote_path)
                except Exception as e:
                    self.logger.warning(f'mkdir {remote_path}: {e}')

                assert recursive, '{} is a directory but recursive has been set to False'.format(local_path)
                assert self.is_directory(sftp, remote_path), \
                    '{} is not a directory on the remote host'.format(remote_path)

                sftp.chdir(remote_path)
                os.chdir(local_path)

                for path, folders, files in os.walk('.'):
                    try:
                        sftp.mkdir(path)
                    except Exception as e:
                        self.logger.warning(f'mkdir {remote_path}: {e}')

                    for file in files:
                        src = os.path.join(path, file)
                        dst = os.path.join(path, file)
                        self.logger.info('Copying {} to {}'.format(src, dst))
                        sftp.put(src, dst)
            else:
                if self.is_directory(sftp, remote_path):
                    remote_path = os.path.join(remote_path, os.path.basename(local_path))

                sftp.put(local_path, remote_path)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def ls(self, path: str = '.', attrs: bool = False, keep_alive: bool = False, **kwargs) \
            -> Union[List[str], Dict[str, Any]]:
        """
        Return the list of files in a path on a remote server.

        :param path: Remote path (default: current directory).
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param attrs: Set to True if you want to get the full information of each file (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        :return: A list of filenames if ``attrs=False``, otherwise a dictionary ``filename -> {attributes`` if
            ``attrs=True``.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        def get_file_type(st_mode: int) -> str:
            if S_ISDIR(st_mode):
                return 'directory'
            elif S_ISBLK(st_mode):
                return 'block'
            elif S_ISCHR(st_mode):
                return 'device'
            elif S_ISDOOR(st_mode):
                return 'door'
            elif S_ISREG(st_mode):
                return 'file'
            elif S_ISLNK(st_mode):
                return 'link'
            elif S_ISFIFO(st_mode):
                return 'fifo'
            elif S_ISSOCK(st_mode):
                return 'sock'
            else:
                return 'unknown'

        try:
            if attrs:
                return {
                    f.filename: {
                        'filename': f.filename,
                        'longname': f.longname,
                        'attributes': f.attr,
                        'type': get_file_type(f.st_mode),
                        'access_time': datetime.datetime.fromtimestamp(f.st_atime),
                        'modify_time': datetime.datetime.fromtimestamp(f.st_mtime),
                        'uid': f.st_uid,
                        'gid': f.st_gid,
                        'size': f.st_size,
                    }
                    for f in sftp.listdir_attr(path)
                }
            else:
                return sftp.listdir(path)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def rm(self, path: str, keep_alive: bool = False, **kwargs) -> None:
        """
        Remove a file from the server.

        :param path: Remote path to remove.
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.remove(path)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def mv(self, path: str, new_path: str, keep_alive: bool = False, **kwargs) -> None:
        """
        Move/rename a file.

        :param path: Remote path to move/rename.
        :param new_path: Destination path.
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.posix_rename(path, new_path)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def mkdir(self, path: str, mode: int = 0o777, keep_alive: bool = False, **kwargs) -> None:
        """
        Create a directory.

        :param path: Path to be created.
        :param mode: Access permissions (default: 0777).
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.mkdir(path, mode=mode)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def rmdir(self, path: str, keep_alive: bool = False, **kwargs) -> None:
        """
        Remove a directory.

        :param path: Path to be removed.
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.rmdir(path)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def ln(self, src: str, dest: str, keep_alive: bool = False, **kwargs) -> None:
        """
        Create a symbolic link.

        :param src: Source path.
        :param dest: Destination path.
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.symlink(src, dest)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def chmod(self, path: str, mode: int, keep_alive: bool = False, **kwargs) -> None:
        """
        Change the access rights of a path.

        :param path: Path to be modified.
        :param mode: Access permissions (in octal mode).
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.chmod(path, mode=mode)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def chown(self, path: str, uid: int, gid: int, keep_alive: bool = False, **kwargs) -> None:
        """
        Change the owner of a path.

        :param path: Path to be modified.
        :param uid: New user ID.
        :param gid: New group ID.
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.chown(path, uid=uid, gid=gid)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def chdir(self, path: str, keep_alive: bool = False, **kwargs) -> None:
        """
        Change directory to the specified path.

        :param path: Destination path.
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.chdir(path)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def getcwd(self, keep_alive: bool = False, **kwargs) -> str:
        """
        Get the current working directory.

        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            return sftp.getcwd()
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def start_forward_tunnel(self, local_port: int, remote_host: str, remote_port: int, bind_addr: Optional[str] = '',
                             **kwargs):
        """
        Start an SSH forward tunnel, tunnelling <local_port> to <remote_host>:<remote_port>.

        :param local_port: Local port.
        :param remote_host: Remote host.
        :param remote_port: Remote port.
        :param bind_addr: If set, the `local_port` will be bound to this address/subnet (default: '', or 0.0.0.0: any).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        key = local_port, remote_host, remote_port
        if key in self._fwd_tunnels:
            self.logger.info('The tunnel {}:{}:{}:{} is already active'.format(
                bind_addr, local_port, remote_host, remote_port))
            return

        client = self._connect(**kwargs)
        server = forward_tunnel(local_port, remote_host, remote_port, client.get_transport(), bind_addr=bind_addr)
        threading.Thread(target=server.serve_forever, name='sshfwdtun').start()

        self._fwd_tunnels[key] = {
            'client': client,
            'server': server,
            'args': kwargs,
        }

    @action
    def stop_forward_tunnel(self, local_port: int, remote_host: str, remote_port: int):
        """
        Stop an active SSH forward tunnel.

        :param local_port: Local port.
        :param remote_host: Remote host.
        :param remote_port: Remote port.
        """
        key = (local_port, remote_host, remote_port)
        if key not in self._fwd_tunnels:
            self.logger.warning('No such forward tunnel: {}:{}:{}'.format(local_port, remote_host, remote_port))
            return

        server = self._fwd_tunnels[key]['server']
        server.server_close()

        args = self._fwd_tunnels[key]['args']
        host, port, user = self._get_host_port_user(**args)
        self.disconnect(host=host, port=port, user=user)

    @action
    def start_reverse_tunnel(self, server_port: int, remote_host: str, remote_port: int, bind_addr: Optional[str] = '',
                             **kwargs):
        """
        Start an SSH reversed tunnel. <server_port> on the SSH server is forwarded across an SSH session back to the
        local machine, and out to a <remote_host>:<remote_port> reachable from this network.

        :param server_port: Server port.
        :param remote_host: Remote host.
        :param remote_port: Remote port.
        :param bind_addr: If set, the `server_port` will be bound to this address/subnet (default: '', or 0.0.0.0: any).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        key = server_port, remote_host, remote_port
        if key in self._fwd_tunnels:
            self.logger.info('The tunnel {}:{}:{}:{} is already active'.format(
                bind_addr, server_port, remote_host, remote_port))
            return

        client = self._connect(**kwargs)
        server = reverse_tunnel(server_port, remote_host, remote_port, transport=client.get_transport(),
                                bind_addr=bind_addr)

        threading.Thread(target=server, name='sshrevtun').start()

        self._rev_tunnels[key] = {
            'client': client,
            'server': server,
            'args': kwargs,
        }

    @action
    def stop_reverse_tunnel(self, server_port: int, remote_host: str, remote_port: int):
        """
        Stop an active SSH reversed tunnel.

        :param server_port: Server port.
        :param remote_host: Remote host.
        :param remote_port: Remote port.
        """
        key = (server_port, remote_host, remote_port)
        if key not in self._rev_tunnels:
            self.logger.warning('No such reversed tunnel: {}:{}:{}'.format(server_port, remote_host, remote_port))
            return

        close_tunnel(*key)
        args = self._rev_tunnels[key]['args']
        host, port, user = self._get_host_port_user(**args)
        self.disconnect(host=host, port=port, user=user)


# vim:sw=4:ts=4:et:



from typing import Optional

from platypush.message.event import Event


class GotifyEvent(Event):
    """
    Gotify base event.
    """


class GotifyMessageEvent(GotifyEvent):
    """
    Event triggered when a message is received on the Gotify instance.
    """
    def __init__(self, *args,
                 message: str,
                 title: Optional[str] = None,
                 priority: Optional[int] = None,
                 extras: Optional[dict] = None,
                 date: Optional[str] = None,
                 id: Optional[int] = None,
                 appid: Optional[int] = None,
                 **kwargs):
        """
        :param message: Message body.
        :param title: Message title.
        :param priority: Message priority.
        :param extras: Message extra payload.
        :param date: Delivery datetime.
        :param id: Message ID.
        :param appid: ID of the sender application.
        """
        super().__init__(
            *args,
            message=message,
            title=title,
            priority=priority,
            extras=extras,
            date=date,
            id=id,
            appid=appid,
            **kwargs
        )



import datetime
import time
from typing import Optional

import requests
from dateutil.tz import gettz, tzutc

from platypush.message.event.sun import SunriseEvent, SunsetEvent
from platypush.plugins import RunnablePlugin, action
from platypush.schemas.sun import SunEventsSchema


class SunPlugin(RunnablePlugin):
    """
    Plugin to get sunset/sunrise events and info for a certain location.

    Triggers:

        * :class:`platypush.message.event.sun.SunriseEvent` on sunrise.
        * :class:`platypush.message.event.sun.SunsetEvent` on sunset.

    """
    _base_url = 'https://api.sunrise-sunset.org/json'
    _attr_to_event_class = {
        'sunrise': SunriseEvent,
        'sunset': SunsetEvent,
    }

    def __init__(self, latitude: float, longitude: float, **kwargs):
        """
        :param latitude: Default latitude.
        :param longitude: Default longitude.
        """
        super().__init__(**kwargs)
        self.latitude = latitude
        self.longitude = longitude

    def main(self):
        while not self.should_stop():
            # noinspection PyUnresolvedReferences
            next_events = self.get_events().output
            next_events = sorted([
                event_class(latitude=self.latitude, longitude=self.longitude, time=next_events[attr])
                for attr, event_class in self._attr_to_event_class.items()
                if next_events.get(attr)
            ], key=lambda t: t.time)

            for event in next_events:
                # noinspection PyTypeChecker
                dt = datetime.datetime.fromisoformat(event.time)
                while (not self.should_stop()) and (dt > datetime.datetime.now(tz=gettz())):
                    time.sleep(1)

                if dt <= datetime.datetime.now(tz=gettz()):
                    self.bus.post(event)

    @staticmethod
    def _convert_time(t: str) -> datetime.datetime:
        now = datetime.datetime.now().replace(tzinfo=gettz())  # lgtm [py/call-to-non-callable]
        dt = datetime.datetime.strptime(t, '%H:%M:%S %p')
        dt = datetime.datetime(year=now.year, month=now.month, day=now.day,
                               hour=dt.hour, minute=dt.minute, second=dt.second, tzinfo=tzutc())

        if dt < now:
            dt += datetime.timedelta(days=1)
        return datetime.datetime.fromtimestamp(dt.timestamp(), tz=gettz())

    @action
    def get_events(self, latitude: Optional[float] = None, longitude: Optional[float] = None) -> dict:
        """
        Return the next sun events.

        :param latitude: Default latitude override.
        :param longitude: Default longitude override.
        :return: .. schema:: sun.SunEventsSchema
        """
        response = requests.get(self._base_url, params={
            'lat': latitude or self.latitude,
            'lng': longitude or self.longitude,
        }).json().get('results', {})

        schema = SunEventsSchema()
        return schema.dump({
            attr: self._convert_time(t)
            for attr, t in response.items()
            if attr in schema.declared_fields.keys()
        })



import queue
import requests
import threading
from typing import List, Optional, Union

from platypush.plugins import action
from platypush.plugins.switch import SwitchPlugin
from platypush.schemas.switchbot import DeviceSchema, DeviceStatusSchema, SceneSchema


class SwitchbotPlugin(SwitchPlugin):
    """
    Plugin to interact with the devices registered to a Switchbot (https://www.switch-bot.com/) account/hub.

    The difference between this plugin and :class:`platypush.plugins.switchbot.bluetooth.SwitchbotBluetoothPlugin` is
    that the latter acts like a Bluetooth hub/bridge that interacts directly with your Switchbot devices, while this
    plugin requires the devices to be connected to a Switchbot Hub and it controls them through your cloud account.

    In order to use this plugin:

        - Set up a Switchbot Hub and configure your devices through the Switchbot app.
        - Follow the steps on the `Switchbot API repo <https://github.com/OpenWonderLabs/SwitchBotAPI#getting-started>`_
          to get an API token from the app.

    """

    def __init__(self, api_token: str, **kwargs):
        """
        :param api_token: API token (see
            `Getting started with the Switchbot API <https://github.com/OpenWonderLabs/SwitchBotAPI#getting-started>`_).
        """
        super().__init__(**kwargs)
        self._api_token = api_token
        self._devices_by_id = {}
        self._devices_by_name = {}

    @staticmethod
    def _url_for(*args, device=None):
        url = 'https://api.switch-bot.com/v1.0/'
        if device:
            url += f'devices/{device["id"]}/'
        url += '/'.join(args)
        return url

    def _run(self, method: str = 'get', *args, device=None, **kwargs):
        response = getattr(requests, method)(self._url_for(*args, device=device), headers={
            'Authorization': self._api_token,
            'Accept': 'application/json',
            'Content-Type': 'application/json; charset=utf-8',
        }, **kwargs)

        response.raise_for_status()
        response = response.json()
        assert response.get('statusCode') == 100, \
            f'Switchbot API request failed: {response.get("statusCode")}: {response.get("message")}'

        return response.get('body')

    def _get_device(self, device: str, use_cache=True):
        if not use_cache:
            self.devices()

        if device in self._devices_by_id:
            return self._devices_by_id[device]
        if device in self._devices_by_name:
            return self._devices_by_name[device]

        assert use_cache, f'Device not found: {device}'
        return self._get_device(device, use_cache=False)

    @action
    def devices(self) -> List[dict]:
        """
        Get the list of devices associated to the specified Switchbot API account.

        :return: .. schema:: switchbot.DeviceSchema(many=True)
        """
        devices = self._run('get', 'devices')
        devices = [
            DeviceSchema().dump({
                **device,
                'is_virtual': False,
            })
            for device in devices.get('deviceList', [])
        ] + [
            DeviceSchema().dump({
                **device,
                'is_virtual': True,
            })
            for device in devices.get('infraredRemoteList', [])
        ]

        for device in devices:
            self._devices_by_id[device['id']] = device
            self._devices_by_name[device['name']] = device

        return devices

    def _worker(self, q: queue.Queue, method: str = 'get', *args, device: Optional[dict] = None, **kwargs):
        schema = DeviceStatusSchema()
        try:
            if method == 'get' and args and args[0] == 'status' and device and device.get('is_virtual'):
                res = schema.load(device)
            else:
                res = self._run(method, *args, device=device, **kwargs)

            q.put(schema.dump(res))
        except Exception as e:
            self.logger.exception(e)
            q.put(e)

    @action
    def status(self, device: Optional[str] = None) -> Union[dict, List[dict]]:
        """
        Get the status of all the registered devices or of a specific device.

        :param device: Filter by device ID or name.
        :return: .. schema:: switchbot.DeviceStatusSchema(many=True)
        """
        # noinspection PyUnresolvedReferences
        devices = self.devices().output
        if device:
            device_info = self._get_device(device)
            status = {} if device_info['is_virtual'] else self._run('get', 'status', device=device_info)
            return {
                **device_info,
                **status,
            }

        devices_by_id = {dev['id']: dev for dev in devices}
        queues = [queue.Queue()] * len(devices)
        workers = [
            threading.Thread(
                target=self._worker,
                args=(queues[i], 'get', 'status'),
                kwargs={'device': dev}
            )
            for i, dev in enumerate(devices)
        ]

        results = []
        for worker in workers:
            worker.start()

        for q in queues:
            response = q.get()
            if not response:
                continue

            assert not isinstance(response, Exception), str(response)
            results.append({
                **devices_by_id.get(response.get('id'), {}),
                **response,
            })

        for worker in workers:
            worker.join()

        return results

    @action
    def press(self, device: str):
        """
        Send a press-button command to a device.

        :param device: Device name or ID.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={'command': 'press'})

    @action
    def toggle(self, device: str, **kwargs):
        """
        Shortcut for :meth:`.press`.

        :param device: Device name or ID.
        """
        return self.press(device)

    @action
    def on(self, device: str, **kwargs):
        """
        Send a turn-on command to a device

        :param device: Device name or ID.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={'command': 'turnOn'})

    @action
    def off(self, device: str, **kwargs):
        """
        Send a turn-off command to a device

        :param device: Device name or ID.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={'command': 'turnOff'})

    @property
    def switches(self) -> List[dict]:
        # noinspection PyUnresolvedReferences
        return [
            dev for dev in self.status().output if 'on' in dev
        ]

    @action
    def set_curtain_position(self, device: str, position: int):
        """
        Set the position of a curtain device.

        :param device: Device name or ID.
        :param position: An integer between 0 (open) and 100 (closed).
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={
            'command': 'setPosition',
            'commandType': 'command',
            'parameter': f'0,ff,{position}',
        })

    @action
    def set_humidifier_efficiency(self, device: str, efficiency: Union[int, str]):
        """
        Set the nebulization efficiency of a humidifier device.

        :param device: Device name or ID.
        :param efficiency: An integer between 0 (open) and 100 (closed) or `auto`.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={
            'command': 'setMode',
            'commandType': 'command',
            'parameter': efficiency,
        })

    @action
    def set_fan_speed(self, device: str, speed: int):
        """
        Set the speed of a fan.

        :param device: Device name or ID.
        :param speed: Speed between 1 and 4.
        """
        # noinspection PyUnresolvedReferences
        status = self.status(device=device).output
        mode = status.get('mode')
        swing_range = status.get('swing_range')
        return self._run('post', 'commands', device=device, json={
            'command': 'set',
            'commandType': 'command',
            'parameter': ','.join(['on', str(mode), str(speed), str(swing_range)]),
        })

    @action
    def set_fan_mode(self, device: str, mode: int):
        """
        Set the mode of a fan.

        :param device: Device name or ID.
        :param mode: Fan mode (1 or 2).
        """
        # noinspection PyUnresolvedReferences
        status = self.status(device=device).output
        speed = status.get('speed')
        swing_range = status.get('swing_range')
        return self._run('post', 'commands', device=device, json={
            'command': 'set',
            'commandType': 'command',
            'parameter': ','.join(['on', str(mode), str(speed), str(swing_range)]),
        })

    @action
    def set_swing_range(self, device: str, swing_range: int):
        """
        Set the swing range of a fan.

        :param device: Device name or ID.
        :param swing_range: Swing range angle, between 0 and 120.
        """
        # noinspection PyUnresolvedReferences
        status = self.status(device=device).output
        speed = status.get('speed')
        mode = status.get('mode')
        return self._run('post', 'commands', device=device, json={
            'command': 'set',
            'commandType': 'command',
            'parameter': ','.join(['on', str(mode), str(speed), str(swing_range)]),
        })

    @action
    def set_temperature(self, device: str, temperature: float):
        """
        Set the temperature of an air conditioner.

        :param device: Device name or ID.
        :param temperature: Temperature, in Celsius.
        """
        # noinspection PyUnresolvedReferences
        status = self.status(device=device).output
        mode = status.get('mode')
        fan_speed = status.get('fan_speed')
        return self._run('post', 'commands', device=device, json={
            'command': 'setAll',
            'commandType': 'command',
            'parameter': ','.join([str(temperature), str(mode), str(fan_speed), 'on']),
        })

    @action
    def set_ac_mode(self, device: str, mode: int):
        """
        Set the mode of an air conditioner.

        :param device: Device name or ID.
        :param mode: Air conditioner mode. Supported values:

            * 1: ``auto``
            * 2: ``cool``
            * 3: ``dry``
            * 4: ``fan``
            * 5: ``heat``

        """
        # noinspection PyUnresolvedReferences
        status = self.status(device=device).output
        temperature = status.get('temperature')
        fan_speed = status.get('fan_speed')
        return self._run('post', 'commands', device=device, json={
            'command': 'setAll',
            'commandType': 'command',
            'parameter': ','.join([str(temperature), str(mode), str(fan_speed), 'on']),
        })

    @action
    def set_ac_fan_speed(self, device: str, fan_speed: int):
        """
        Set the fan speed for an air conditioner.

        :param device: Device name or ID.
        :param fan_speed: Possible values:

            * 1: ``auto``
            * 2: ``low``
            * 3: ``medium``
            * 4: ``high``

        """
        # noinspection PyUnresolvedReferences
        status = self.status(device=device).output
        temperature = status.get('temperature')
        mode = status.get('mode')
        return self._run('post', 'commands', device=device, json={
            'command': 'setAll',
            'commandType': 'command',
            'parameter': ','.join([str(temperature), str(mode), str(fan_speed), 'on']),
        })

    @action
    def set_channel(self, device: str, channel: int):
        """
        Set the channel on a TV, IPTV/Streamer, Set Top Box device.

        :param device: Device name or ID.
        :param channel: Channel number.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={
            'command': 'SetChannel',
            'commandType': 'command',
            'parameter': [str(channel)],
        })

    @action
    def volup(self, device: str):
        """
        Send volume up IR event to a device (for TV, IPTV/Streamer, Set Top Box, DVD and Speaker).

        :param device: Device name or ID.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={
            'command': 'volumeAdd',
            'commandType': 'command',
        })

    @action
    def voldown(self, device: str):
        """
        Send volume down IR event to a device (for TV, IPTV/Streamer, Set Top Box, DVD and Speaker).

        :param device: Device name or ID.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={
            'command': 'volumeSub',
            'commandType': 'command',
        })

    @action
    def mute(self, device: str):
        """
        Send mute/unmute IR event to a device (for TV, IPTV/Streamer, Set Top Box, DVD and Speaker).

        :param device: Device name or ID.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={
            'command': 'setMute',
            'commandType': 'command',
        })

    @action
    def channel_next(self, device: str):
        """
        Send next channel IR event to a device (for TV, IPTV/Streamer, and Set Top Box).

        :param device: Device name or ID.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={
            'command': 'channelAdd',
            'commandType': 'command',
        })

    @action
    def channel_prev(self, device: str):
        """
        Send previous channel IR event to a device (for TV, IPTV/Streamer, and Set Top Box).

        :param device: Device name or ID.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={
            'command': 'channelSub',
            'commandType': 'command',
        })

    @action
    def play(self, device: str):
        """
        Send play IR event to a device (for DVD and Speaker).

        :param device: Device name or ID.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={
            'command': 'Play',
            'commandType': 'command',
        })

    @action
    def pause(self, device: str):
        """
        Send pause IR event to a device (for DVD and Speaker).

        :param device: Device name or ID.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={
            'command': 'Pause',
            'commandType': 'command',
        })

    @action
    def stop(self, device: str):
        """
        Send stop IR event to a device (for DVD and Speaker).

        :param device: Device name or ID.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={
            'command': 'Stop',
            'commandType': 'command',
        })

    @action
    def forward(self, device: str):
        """
        Send forward IR event to a device (for DVD and Speaker).

        :param device: Device name or ID.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={
            'command': 'FastForward',
            'commandType': 'command',
        })

    @action
    def back(self, device: str):
        """
        Send backward IR event to a device (for DVD and Speaker).

        :param device: Device name or ID.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={
            'command': 'Rewind',
            'commandType': 'command',
        })

    @action
    def next(self, device: str):
        """
        Send next IR event to a device (for DVD and Speaker).

        :param device: Device name or ID.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={
            'command': 'Next',
            'commandType': 'command',
        })

    @action
    def previous(self, device: str):
        """
        Send previous IR event to a device (for DVD and Speaker).

        :param device: Device name or ID.
        """
        device = self._get_device(device)
        return self._run('post', 'commands', device=device, json={
            'command': 'Previous',
            'commandType': 'command',
        })

    @action
    def scenes(self) -> List[dict]:
        """
        Get the list of registered scenes.

        :return: .. schema:: switchbot.SceneSchema(many=True)
        """
        return SceneSchema().dump(self._run('get', 'scenes'), many=True)

    @action
    def run_scene(self, scene: str):
        """
        Execute a scene.

        :param scene: Scene ID or name.
        """
        # noinspection PyUnresolvedReferences
        scenes = [
            s for s in self.scenes().output
            if s.get('id') == scene or s.get('name') == scene
        ]

        assert scenes, f'No such scene: {scene}'
        return self._run('post', 'scenes', scenes[0]['id'], 'execute')


# vim:sw=4:ts=4:et:



from platypush.backend.weather import WeatherBackend


class WeatherDarkskyBackend(WeatherBackend):
    """
    Weather forecast backend that leverages the DarkSky API.

    Triggers:

        * :class:`platypush.message.event.weather.NewWeatherConditionEvent` when there is a weather condition update

    Requires:

        * The :class:`platypush.plugins.weather.darksky.WeatherDarkskyPlugin` plugin configured

    """

    def __init__(self, poll_seconds: int = 300, **kwargs):
        """
        :param poll_seconds: How often the backend should check for updates (default: every 5 minutes).
        """
        super().__init__(plugin_name='weather.darksky', poll_seconds=poll_seconds, **kwargs)


# vim:sw=4:ts=4:et:



import contextlib
import logging
import multiprocessing
import os
import pathlib
import re
import select
import socket
import ssl as _ssl
import struct
from typing import Iterable, Optional, Type, Sequence, Dict, Tuple

import irc.bot
from irc.client import Connection, Event as _IRCEvent, ServerConnection, DCCConnection, ip_numstr_to_quad, \
    ip_quad_to_numstr
from irc.connection import Factory as ConnectionFactory
from irc.events import codes as irc_event_codes

from platypush.context import get_bus
from platypush.message.event.irc import IRCEvent, IRCChannelJoinEvent, IRCChannelKickEvent, IRCDisconnectEvent, \
    IRCModeEvent, IRCNickChangeEvent, IRCPartEvent, IRCQuitEvent, IRCConnectEvent, IRCPrivateMessageEvent, \
    IRCPublicMessageEvent, IRCDCCRequestEvent, IRCDCCMessageEvent, IRCDCCFileRequestEvent, IRCCTCPMessageEvent, \
    IRCDCCFileRecvCompletedEvent, IRCDCCFileRecvCancelledEvent, IRCDCCFileSendCompletedEvent, \
    IRCDCCFileSendCancelledEvent


class IRCBot(irc.bot.SingleServerIRCBot):
    def __init__(
            self, server: str, port: int, nickname: str, alias: str,
            channels: Iterable[str],
            response_timeout: Optional[float],
            dcc_file_transfer_timeout: Optional[float],
            dcc_accept_timeout: Optional[float],
            dcc_downloads_dir: str, realname: Optional[str] = None,
            password: Optional[str] = None, ssl: bool = False,
            ipv6: bool = False, stop_message: Optional[str] = None,
            dcc_ip_whitelist: Optional[Sequence[str]] = None,
            dcc_ip_blacklist: Optional[Sequence[str]] = None,
            dcc_nick_whitelist: Optional[Sequence[str]] = None,
            dcc_nick_blacklist: Optional[Sequence[str]] = None,
            dcc_max_connections: Optional[int] = None,
    ):
        connection_factory = ConnectionFactory()
        if ssl:
            connection_factory.wrapper = _ssl.wrap_socket
        if ipv6:
            connection_factory.family = socket.AF_INET6

        super().__init__(
            server_list=[(server, port)], nickname=nickname,
            realname=realname, connect_factory=connection_factory
        )

        self.server = server
        self.port = port
        self.alias = alias
        self._password = password
        self.channels.update({channel: None for channel in channels})
        self._stop_message = stop_message
        self.dcc_ip_whitelist = set(dcc_ip_whitelist or [])
        self.dcc_ip_blacklist = set(dcc_ip_blacklist or [])
        self.dcc_nick_whitelist = set(dcc_nick_whitelist or [])
        self.dcc_nick_blacklist = set(dcc_nick_blacklist or [])
        self.dcc_downloads_dir = dcc_downloads_dir
        self.response_timeout = response_timeout
        self.dcc_file_transfer_timeout = dcc_file_transfer_timeout
        self.dcc_accept_timeout = dcc_accept_timeout
        self.dcc_max_connections = dcc_max_connections
        self._dcc_send_procs: Dict[Tuple[str, int], multiprocessing.Process] = {}
        self._dcc_recv_procs: Dict[Tuple[str, int], multiprocessing.Process] = {}
        self._dcc_proc_completion_queue = multiprocessing.Queue()
        self._dcc_processor: Optional[multiprocessing.Process] = None
        self.logger = logging.getLogger(f'irc@{server}')
        # Maps <matching_event_type> -> <response_queue>
        self._pending_requests: Dict[str, multiprocessing.Queue] = {}

    @property
    def stop_message(self) -> Optional[str]:
        return self._stop_message

    def _post_event(self, connection: Connection, output_event_type: Type[IRCEvent], **kwargs):
        if isinstance(connection, ServerConnection):
            kwargs['server'] = connection.server
            kwargs['port'] = connection.port

        kwargs['connected'] = connection.connected
        kwargs['alias'] = self.alias
        event = output_event_type(**kwargs)
        get_bus().post(event)

    def on_join(self, connection: ServerConnection, event: _IRCEvent):
        self._post_event(connection, IRCChannelJoinEvent, channel=event.target, nick=event.source.nick)
        super()._on_join(connection, event)

    def on_kick(self, connection: ServerConnection, event: _IRCEvent):
        self._post_event(
            connection, IRCChannelKickEvent, channel=event.target,
            source_nick=event.arguments[1], target_nick=event.arguments[0]
        )

        super()._on_kick(connection, event)

    def on_nick(self, connection: ServerConnection, event: _IRCEvent):
        self._post_event(
            connection, IRCNickChangeEvent, before=event.source.nick, after=event.target.nick
        )
        super()._on_nick(connection, event)

    def on_mode(self, connection: ServerConnection, event: _IRCEvent):
        self._post_event(
            connection, IRCModeEvent, mode=event.arguments[0],
            source=event.source.nick,
            target_=(event.arguments[1] if len(event.arguments) > 1 else None),
            channel=event.target,
        )

        super()._on_mode(connection, event)

    def on_part(self, connection: Connection, event: _IRCEvent):
        self._post_event(connection, IRCPartEvent, nick=event.source.nick)
        super()._on_part(connection, event)

    def on_quit(self, connection: Connection, event: _IRCEvent):
        self._post_event(connection, IRCQuitEvent, nick=event.source.nick)
        super()._on_quit(connection, event)

    def on_welcome(self, connection: Connection, *_):
        self._post_event(connection, IRCConnectEvent)
        for channel in self.channels:
            self.logger.info(f'Joining channel {channel}')
            self.connection.join(channel)

    @staticmethod
    def on_nicknameinuse(connection: ServerConnection, *_):
        connection.nick(connection.nickname + '_')

    @staticmethod
    def _mentions_me(connection: ServerConnection, text: str) -> bool:
        return bool(re.search(fr'(^|\s|@){connection.nickname}(:|\s|!|\?|$)', text))

    def on_pubmsg(self, connection: ServerConnection, event: _IRCEvent):
        self._post_event(
            connection, IRCPublicMessageEvent,
            text=event.arguments[0], nick=event.source.nick,
            channel=event.target,
            mentions_me=self._mentions_me(connection, event.arguments[0])
        )

    def on_privmsg(self, connection: ServerConnection, event: _IRCEvent):
        self._post_event(
            connection, IRCPrivateMessageEvent,
            text=event.arguments[0], nick=event.source.nick,
            channel=event.target,
            mentions_me=self._mentions_me(connection, event.arguments[0])
        )

    def on_dccchat(self, connection: DCCConnection, event: _IRCEvent):
        if len(event.arguments) != 2:
            return

        args = event.arguments[1].split()
        if len(args) == 4:
            try:
                address = ip_numstr_to_quad(args[2])
                port = int(args[3])
            except ValueError:
                return

            nick = event.source.nick
            if not self._is_dcc_connection_request_allowed(address=address, nick=nick):
                self.logger.info(f'Refused DCC connection from address={address} nick={nick}')
                connection.disconnect('Unauthorized peer')
                return

            self._post_event(
                connection, IRCDCCRequestEvent, address=address, port=port, nick=nick
            )

            self.dcc('chat').connect(address, port)

    def _read(self, sock: socket.socket, bufsize: int = 4096) -> bytes:
        if self.dcc_file_transfer_timeout:
            sock.setblocking(False)
            has_data = select.select([sock], [], [], self.dcc_file_transfer_timeout)
            if has_data[0]:
                return sock.recv(bufsize)
            raise TimeoutError(f'Time out ({self.dcc_file_transfer_timeout}s)')

        return sock.recv(bufsize)

    def _dcc_connect_processor(self):
        while True:
            proc_type, addr, port = self._dcc_proc_completion_queue.get()
            conn_map = getattr(self, f'_dcc_{proc_type}_procs', {})
            conn_map.pop((addr, port), None)

    def _process_dcc_recv(self, connection: DCCConnection, filename: str, address: str, port: int, size: int):
        pathlib.Path(self.dcc_downloads_dir).mkdir(parents=True, exist_ok=True)
        filename = os.path.abspath(os.path.join(self.dcc_downloads_dir, filename))
        assert filename.startswith(self.dcc_downloads_dir), (
            'Attempt to save a file outside the downloads directory'
        )

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock, \
                open(filename, 'wb') as f:
            try:
                sock.connect((address, port))
                processed_bytes = 0

                while processed_bytes < size:
                    buf = self._read(sock)
                    processed_bytes += len(buf)
                    f.write(buf)
                    sock.send(f'{socket.htonl(processed_bytes)}'.encode())

                self._post_event(
                    connection, IRCDCCFileRecvCompletedEvent, address=address,
                    port=port, file=filename, size=size
                )
            except Exception as e:
                self.logger.error(f'DCC transfer error from {address}:{port}: {e}')
                self.logger.exception(e)
                self._post_event(
                    connection, IRCDCCFileRecvCancelledEvent, address=address,
                    port=port, file=filename, error=str(e)
                )
            finally:
                connection.disconnect()
                self._dcc_proc_completion_queue.put(('recv', address, port))

    def _process_dcc_send(self, connection: DCCConnection, filename: str):
        try:
            dcc_sock, addr = connection.socket.accept()
            self._set_accept_timeout(dcc_sock)
            self.logger.info(f'Accepted DCC connection from {addr}')

            # Create a new server socket for the file transfer
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as srv_sock:
                srv_sock.bind((connection.localaddress, 0))
                address, transfer_port = srv_sock.getsockname()
                dcc_sock.send(
                    (
                        f'CTCP_MESSAGE \x01'
                        f'DCC SEND {os.path.basename(filename)} '
                        f'{ip_quad_to_numstr(address)} {transfer_port} '
                        f'{os.path.getsize(filename)}\x01\r\n'
                    ).encode()
                )

                # Wait for a connection on the file transfer port
                self._set_accept_timeout(srv_sock)
                srv_sock.listen(1)
                client_sock, addr = srv_sock.accept()
                self.logger.info(f'DCC file transfer for {filename} accepted on {addr}')

                try:
                    with open(filename, 'rb') as f:
                        buf = f.read(4096)
                        while buf:
                            client_sock.send(buf)
                            recv_bytes = struct.unpack('>i', self._read(client_sock))[0]
                            buf = buf[recv_bytes:]

                            if not buf:
                                buf = f.read(4096)

                    self._post_event(
                        connection, IRCDCCFileSendCompletedEvent,
                        file=filename, address=address, port=transfer_port
                    )
                finally:
                    client_sock.close()
                    dcc_sock.close()
        except Exception as e:
            self.logger.error(f'DCC transfer error to {connection.peeraddress}:{connection.peerport}: {e}')
            self.logger.exception(e)
            self._post_event(
                connection, IRCDCCFileSendCancelledEvent, address=connection.peeraddress,
                port=connection.peerport, file=filename, error=str(e)
            )
        finally:
            connection.disconnect()
            self._dcc_proc_completion_queue.put(('send', connection.localaddress, connection.localport))

    def _set_accept_timeout(self, sock: socket.socket):
        if self.dcc_accept_timeout:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.settimeout(self.dcc_accept_timeout)

    def _stop_dcc_connection(self, address: str, port: int):
        self.logger.info(f'Stopping DCC connection to {address}:{port}')
        proc = self._dcc_recv_procs.pop((address, port), None)
        if proc and proc.is_alive():
            proc.terminate()

    def _accept_dcc_file_request(self, connection: DCCConnection, ctcp_msg: str, nick: str):
        ctcp_msg = [
            token.strip()
            for token in ctcp_msg.split(' ')
            if token
        ]

        filename = ' '.join(ctcp_msg[2:-3])
        address, port, size = ctcp_msg[-3:]

        if not (filename and address and port):
            return

        address = ip_numstr_to_quad(address)
        port = int(port)
        size = int(size)

        if port == 0:
            self.logger.warning('Passive CTCP file transfer is currently not supported')
            return

        self._post_event(
            connection, IRCDCCFileRequestEvent,
            address=address,
            port=port, file=filename,
            size=size, nick=nick
        )

        # Accept the file request
        # (if we're here then the peer is whitelisted/not blacklisted)
        self._dcc_recv_procs[(address, port)] = multiprocessing.Process(
            target=self._process_dcc_recv, kwargs={
                'connection': connection,
                'address': address,
                'port': port,
                'filename': filename,
                'size': size,
            }
        )

        self._dcc_recv_procs[(address, port)].start()

    def _handle_ctcp_message(self, connection: DCCConnection, event: _IRCEvent):
        body = event.arguments[0]
        sep_pos = [i for i, ch in enumerate(body) if ch == ord(b'\x01')]
        msg = []

        try:
            msg = [
                body[(0 if i == 0 else sep_pos[i - 1] + 1):pos].decode().strip()
                for i, pos in enumerate(sep_pos)
            ][1:]
        except (ValueError, TypeError):
            pass

        if not msg:
            return

        # Check if this is a DCC file send request
        if msg[0].startswith('DCC SEND '):
            self._accept_dcc_file_request(connection, msg[0], event.source)
        else:
            self._post_event(
                connection, IRCCTCPMessageEvent, address=event.source, message=msg
            )

    def on_dccmsg(self, connection: DCCConnection, event: _IRCEvent):
        ctcp_header = b'CTCP_MESSAGE'
        if event.arguments[0][:len(ctcp_header)] == ctcp_header:
            return self._handle_ctcp_message(connection, event)

        self._post_event(connection, IRCDCCMessageEvent, address=event.source, body=event.arguments[0])

    def on_disconnect(self, connection: Connection, event: _IRCEvent):
        self._post_event(connection, IRCDisconnectEvent)
        # Cache channels for reconnect logic
        channels = {ch: None for ch in self.channels.keys()}
        super()._on_disconnect(connection, event)
        self.channels.update(channels)

    def on_whoisuser(self, _, event: _IRCEvent):
        self._process_event_on_queue(event)

    def on_currenttopic(self, _, event: _IRCEvent):
        self._process_event_on_queue(event)

    def _process_event_on_queue(self, event: _IRCEvent):
        q = self._pending_requests.get(event.type)
        if q:
            q.put(event)

    def _on_generic_event(self, _, event: _IRCEvent):
        self.logger.debug(f'Received raw unhandled IRC event on {self.server}: {event.__dict__}')

    def _is_dcc_connection_request_allowed(self, address: str, nick: str) -> bool:
        if self.dcc_ip_whitelist and address not in self.dcc_ip_whitelist:
            return False
        if self.dcc_ip_blacklist and address in self.dcc_ip_blacklist:
            return False
        if self.dcc_nick_whitelist and nick not in self.dcc_nick_whitelist:
            return False
        if self.dcc_nick_blacklist and nick in self.dcc_nick_blacklist:
            return False

        if self.dcc_max_connections and len(self._dcc_recv_procs) >= self.dcc_max_connections:
            self.logger.warning(
                f'Refused new DCC connection: maximum number of concurrent '
                f'connections ({self.dcc_max_connections}) reached'
            )

            return False

        return True

    def dcc_file_transfer(self, file: str, nick: str, bind_address: Optional[str] = None):
        conn: DCCConnection = self.dcc('chat')
        bind_address = (bind_address, 0) if bind_address else None
        conn.listen(bind_address)
        conn.passive = False
        self.connection.privmsg(
            nick,
            f'\x01DCC CHAT chat {ip_quad_to_numstr(conn.localaddress)} {conn.localport}\x01'
        )

        send_proc = self._dcc_send_procs[(conn.localaddress, conn.localport)] = multiprocessing.Process(
            target=self._process_dcc_send,
            kwargs={
                'connection': conn,
                'filename': file,
            }
        )

        send_proc.start()
        send_proc.join()

    @contextlib.contextmanager
    def event_queue(self, event_type: str) -> multiprocessing.Queue:
        q = self._pending_requests[event_type] = multiprocessing.Queue()
        try:
            yield q
        finally:
            q.close()
            self._pending_requests.pop(event_type, None)

    def start(self):
        self._dcc_processor = multiprocessing.Process(target=self._dcc_connect_processor)
        self._dcc_processor.start()

        for event_code in irc_event_codes.keys():
            handler = getattr(self, f'on_{event_code}', None)
            if not handler:
                self.reactor.add_global_handler(
                    event_code, lambda conn, evt: self._on_generic_event(conn, evt)
                )

        super().start()

    def stop(self, msg: Optional[str] = None):
        msg = msg or self.stop_message
        for addr, port in list(self._dcc_recv_procs.keys()) + list(self._dcc_send_procs.keys()):
            self._stop_dcc_connection(addr, port)

        if self._dcc_processor and self._dcc_processor.is_alive():
            self._dcc_processor.terminate()
        self._dcc_processor = None

        super().disconnect(msg=msg)



import io
import os
import pathlib
import socket
import threading
import time

from abc import ABC, abstractmethod
from contextlib import contextmanager
from datetime import datetime
from multiprocessing import Process
from queue import Queue
from typing import Optional, Union, Dict, Tuple, IO

from platypush.config import Config
from platypush.message.event.camera import CameraRecordingStartedEvent, CameraPictureTakenEvent, \
    CameraRecordingStoppedEvent, CameraVideoRenderedEvent
from platypush.plugins import Plugin, action
from platypush.plugins.camera.model.camera import CameraInfo, Camera
from platypush.plugins.camera.model.exceptions import CameraException, CaptureAlreadyRunningException
from platypush.plugins.camera.model.writer import VideoWriter, StreamWriter
from platypush.plugins.camera.model.writer.ffmpeg import FFmpegFileWriter
from platypush.plugins.camera.model.writer.preview import PreviewWriter, PreviewWriterFactory
from platypush.utils import get_plugin_name_by_class

__all__ = ['Camera', 'CameraInfo', 'CameraException', 'CameraPlugin', 'CaptureAlreadyRunningException',
           'VideoWriter', 'StreamWriter', 'PreviewWriter']


class CameraPlugin(Plugin, ABC):
    """
    Abstract plugin to control camera devices.

    If the :class:`platypush.backend.http.HttpBackend` is enabled then the plugins that implement this class can
    expose two endpoints:

        - ``http://host:8008/camera/<plugin>/photo<.extension>`` to capture a photo from the camera, where
            ``.extension`` can be ``.jpg``, ``.png`` or ``.bmp``.
        - ``http://host:8008/camera/<plugin>/video<.extension>`` to get a live feed from the camera, where
            ``.extension`` can be ``.mjpeg``, ``.mkv``/``.webm``, ``.mp4``/``.h264`` or ``.h265``.

    Both the endpoints support the same parameters of the constructor of this class (e.g. ``device``, ``warmup_frames``,
    ``duration`` etc.) as ``GET`` parameters.

    Requires:

        * **Pillow** (``pip install Pillow``) [optional] default handler for image transformations.
        * **wxPython** (``pip install wxPython``) [optional] default handler for camera previews (``ffplay`` will be
            used as a fallback if ``wxPython`` is not installed).
        * **ffmpeg** (see installation instructions for your OS) for rendering/streaming videos.

    Triggers:

        * :class:`platypush.message.event.camera.CameraRecordingStartedEvent`
            when a new video recording/photo burst starts
        * :class:`platypush.message.event.camera.CameraRecordingStoppedEvent`
            when a video recording/photo burst ends
        * :class:`platypush.message.event.camera.CameraVideoRenderedEvent`
            when a sequence of captured is successfully rendered into a video
        * :class:`platypush.message.event.camera.CameraPictureTakenEvent`
            when a snapshot is captured and stored to an image file

    """

    _camera_class = Camera
    _camera_info_class = CameraInfo
    _video_writer_class = FFmpegFileWriter

    def __init__(self, device: Optional[Union[int, str]] = None, resolution: Tuple[int, int] = (640, 480),
                 frames_dir: Optional[str] = None, warmup_frames: int = 5, warmup_seconds: Optional[float] = 0.,
                 capture_timeout: Optional[float] = 20.0, scale_x: Optional[float] = None,
                 scale_y: Optional[float] = None, rotate: Optional[float] = None, grayscale: Optional[bool] = None,
                 color_transform: Optional[Union[int, str]] = None, fps: float = 16, horizontal_flip: bool = False,
                 vertical_flip: bool = False, input_format: Optional[str] = None, output_format: Optional[str] = None,
                 stream_format: str = 'mjpeg', listen_port: Optional[int] = 5000, bind_address: str = '0.0.0.0',
                 ffmpeg_bin: str = 'ffmpeg', input_codec: Optional[str] = None, output_codec: Optional[str] = None,
                 **kwargs):
        """
        :param device: Identifier of the default capturing device.
        :param resolution: Default resolution, as a tuple of two integers.
        :param frames_dir: Directory where the camera frames will be stored (default:
            ``~/.local/share/platypush/<plugin.name>/frames``)
        :param warmup_frames: Cameras usually take a while to adapt their
            luminosity and focus to the environment when taking a picture.
            This parameter allows you to specify the number of "warmup" frames
            to capture upon picture command before actually capturing a frame
            (default: 5 but you may want to calibrate this parameter for your
            camera)
        :param warmup_seconds: Number of seconds to wait before a picture is taken or the first frame of a
            video/sequence is captured (default: 0).
        :param capture_timeout: Maximum number of seconds to wait between the programmed termination of a capture
            session and the moment the device is released.
        :param scale_x: If set, the images will be scaled along the x axis by the specified factor
        :param scale_y: If set, the images will be scaled along the y axis by the specified factor
        :param color_transform: Color transformation to apply to the images.
        :param grayscale: Whether the output should be converted to grayscale.
        :param rotate: If set, the images will be rotated by the specified number of degrees
        :param fps: Frames per second (default: 25).
        :param horizontal_flip: If set, the images will be flipped on the horizontal axis.
        :param vertical_flip: If set, the images will be flipped on the vertical axis.
        :param listen_port: Default port to be used for streaming over TCP (default: 5000).
        :param bind_address: Default bind address for TCP streaming (default: 0.0.0.0, accept any connections).
        :param input_codec: Specify the ffmpeg video codec (``-vcodec``) used for the input.
        :param output_codec: Specify the ffmpeg video codec (``-vcodec``) to be used for encoding the output. For some
            ffmpeg output formats (e.g. ``h264`` and ``rtp``) this may default to ``libxvid``.
        :param input_format: Plugin-specific format/type for the input stream.
        :param output_format: Plugin-specific format/type for the output videos.
        :param ffmpeg_bin: Path to the ffmpeg binary (default: ``ffmpeg``).
        :param stream_format: Default format for the output when streamed to a network device. Available:

            - ``MJPEG`` (default)
            - ``H264`` (over ``ffmpeg``)
            - ``H265`` (over ``ffmpeg``)
            - ``MKV`` (over ``ffmpeg``)
            - ``MP4`` (over ``ffmpeg``)

        """
        super().__init__(**kwargs)

        self.workdir = os.path.join(Config.get('workdir'), get_plugin_name_by_class(self))
        pathlib.Path(self.workdir).mkdir(mode=0o755, exist_ok=True, parents=True)

        # noinspection PyArgumentList
        self.camera_info = self._camera_info_class(device, color_transform=color_transform, warmup_frames=warmup_frames,
                                                   warmup_seconds=warmup_seconds, rotate=rotate, scale_x=scale_x,
                                                   scale_y=scale_y, capture_timeout=capture_timeout, fps=fps,
                                                   input_format=input_format, output_format=output_format,
                                                   stream_format=stream_format, resolution=resolution,
                                                   grayscale=grayscale, listen_port=listen_port,
                                                   horizontal_flip=horizontal_flip, vertical_flip=vertical_flip,
                                                   ffmpeg_bin=ffmpeg_bin, input_codec=input_codec,
                                                   output_codec=output_codec, bind_address=bind_address,
                                                   frames_dir=os.path.abspath(
                                                       os.path.expanduser(frames_dir or
                                                                          os.path.join(self.workdir, 'frames'))))

        self._devices: Dict[Union[int, str], Camera] = {}
        self._streams: Dict[Union[int, str], Camera] = {}

    def _merge_info(self, **info) -> CameraInfo:
        merged_info = self.camera_info.clone()
        merged_info.set(**info)
        return merged_info

    def open_device(self, device: Optional[Union[int, str]] = None, stream: bool = False, **params) -> Camera:
        """
        Initialize and open a device.

        :return: The initialized camera device.
        :raises: :class:`platypush.plugins.camera.CaptureSessionAlreadyRunningException`
        """
        info = self._merge_info(**params)
        if device is None:
            device = info.device
        elif device not in self._devices:
            info.device = device
        else:
            info = self._devices[device].info.clone()

        assert device is not None, 'No device specified/configured'
        if device in self._devices:
            camera = self._devices[device]
            if camera.capture_thread and camera.capture_thread.is_alive() and camera.start_event.is_set():
                raise CaptureAlreadyRunningException(device)

            camera.start_event.clear()
            camera.capture_thread = None
        else:
            # noinspection PyArgumentList
            camera = self._camera_class(info=info)

        camera.info.set(**params)
        camera.object = self.prepare_device(camera)

        if stream:
            writer_class = StreamWriter.get_class_by_name(camera.info.stream_format)
            camera.stream = writer_class(camera=camera, plugin=self)

        if camera.info.frames_dir:
            pathlib.Path(os.path.abspath(os.path.expanduser(camera.info.frames_dir))).mkdir(
                mode=0o755, exist_ok=True, parents=True)

        self._devices[device] = camera
        return camera

    def close_device(self, camera: Camera, wait_capture: bool = True) -> None:
        """
        Close and release a device.
        """
        name = camera.info.device
        self.stop_preview(camera)
        self.release_device(camera)

        camera.start_event.clear()
        if wait_capture:
            self.wait_capture(camera)

        if name in self._devices:
            del self._devices[name]

    def wait_capture(self, camera: Camera) -> None:
        """
        Wait until a capture session terminates.

        :param camera: Camera object. ``camera.info.capture_timeout`` is used as a capture thread termination timeout
            if set.
        """
        if camera.capture_thread and camera.capture_thread.is_alive() and \
                threading.get_ident() != camera.capture_thread.ident:
            try:
                camera.capture_thread.join(timeout=camera.info.capture_timeout)
            except Exception as e:
                self.logger.warning('Error on FFmpeg capture wait: {}'.format(str(e)))

    @contextmanager
    def open(self, device: Optional[Union[int, str]] = None, stream: bool = None, **info) -> Camera:
        """
        Initialize and open a device using a context manager pattern.

        :param device: Capture device by name, path or ID.
        :param stream: If set, the frames will be streamed to ``camera.stream``.
        :param info: Camera parameters override - see constructors parameters.
        :return: The initialized :class:`platypush.plugins.camera.Camera` object.
        """
        camera = None
        try:
            camera = self.open_device(device, stream=stream, **info)
            yield camera
        finally:
            self.close_device(camera)

    @abstractmethod
    def prepare_device(self, device: Camera):
        """
        Prepare a device using the plugin-specific logic - to be implemented by the derived classes.

        :param device: An initialized :class:`platypush.plugins.camera.Camera` object.
        """
        raise NotImplementedError()

    @abstractmethod
    def release_device(self, device: Camera):
        """
        Release a device using the plugin-specific logic - to be implemented by the derived classes.

        :param device: An initialized :class:`platypush.plugins.camera.Camera` object.
        """
        raise NotImplementedError()

    @abstractmethod
    def capture_frame(self, device: Camera, *args, **kwargs):
        """
        Capture a frame from a device using the plugin-specific logic - to be implemented by the derived classes.

        :param device: An initialized :class:`platypush.plugins.camera.Camera` object.
        """
        raise NotImplementedError()

    # noinspection PyShadowingBuiltins
    @staticmethod
    def store_frame(frame, filepath: str, format: Optional[str] = None):
        """
        Capture a frame to the filesystem using the ``PIL`` library - it can be overridden by derived classes.

        :param frame: Frame object (default: a byte-encoded object or a ``PIL.Image`` object).
        :param filepath: Destination file.
        :param format: Output format.
        """
        from PIL import Image
        if isinstance(frame, bytes):
            frame = list(frame)
        elif not isinstance(frame, Image.Image):
            frame = Image.fromarray(frame)

        save_args = {}
        if format:
            save_args['format'] = format

        frame.save(filepath, **save_args)

    def _store_frame(self, frame, frames_dir: Optional[str] = None, image_file: Optional[str] = None,
                     *args, **kwargs) -> str:
        """
        :meth:`.store_frame` wrapper.
        """
        if image_file:
            filepath = os.path.abspath(os.path.expanduser(image_file))
        else:
            filepath = os.path.abspath(os.path.expanduser(
                os.path.join(frames_dir or '', datetime.now().strftime('%Y-%m-%d_%H-%M-%S-%f.jpg'))))

        pathlib.Path(filepath).parent.mkdir(mode=0o755, exist_ok=True, parents=True)
        self.store_frame(frame, filepath, *args, **kwargs)
        return filepath

    def start_preview(self, camera: Camera):
        if camera.preview and not camera.preview.closed:
            self.logger.info('A preview window is already active on device {}'.format(camera.info.device))
            return

        camera.preview = PreviewWriterFactory.get(camera, self)
        if isinstance(camera.preview, Process):
            camera.preview.start()

    def stop_preview(self, camera: Camera):
        if camera.preview and not camera.preview.closed:
            camera.preview.close()

        if isinstance(camera.preview, Process) and camera.preview.is_alive():
            camera.preview.terminate()
            camera.preview.join(timeout=5.0)

        if isinstance(camera.preview, Process) and camera.preview.is_alive():
            camera.preview.kill()

        camera.preview = None

    def frame_processor(self, frame_queue: Queue, camera: Camera, image_file: Optional[str] = None):
        while True:
            frame = frame_queue.get()
            if frame is None:
                break

            frame = self.transform_frame(frame, camera.info.color_transform)
            if camera.info.grayscale:
                frame = self.to_grayscale(frame)

            frame = self.rotate_frame(frame, camera.info.rotate)
            frame = self.flip_frame(frame, camera.info.horizontal_flip, camera.info.vertical_flip)
            frame = self.scale_frame(frame, camera.info.scale_x, camera.info.scale_y)

            for output in camera.get_outputs():
                output.write(frame)

            if camera.info.frames_dir or image_file:
                self._store_frame(frame=frame, frames_dir=camera.info.frames_dir, image_file=image_file)

    def capturing_thread(self, camera: Camera, duration: Optional[float] = None, video_file: Optional[str] = None,
                         image_file: Optional[str] = None, n_frames: Optional[int] = None, preview: bool = False,
                         **kwargs):
        """
        Camera capturing thread.

        :param camera: An initialized :class:`platypush.plugins.camera.Camera` object.
        :param duration: Capturing session duration in seconds (default: until :meth:`.stop_capture` is called).
        :param video_file: If set, the session will be recorded to this output video file (video capture mode).
        :param image_file: If set, the output of the session will be a single image file (photo mode).
        :param n_frames: Number of frames to be captured (default: until :meth:`.stop_capture` is called).
        :param preview: Start a preview window.
        :param kwargs: Extra arguments to be passed to :meth:`.capture_frame`.
        """
        camera.start_event.wait()
        recording_started_time = time.time()
        captured_frames = 0

        evt_args = {
            'device': camera.info.device,
        }

        if video_file or image_file:
            evt_args['filename'] = video_file or image_file
        if camera.info.frames_dir:
            evt_args['frames_dir'] = camera.info.frames_dir
        if preview:
            self.start_preview(camera)
        if duration and camera.info.warmup_seconds:
            duration = duration + camera.info.warmup_seconds
        if video_file:
            camera.file_writer = self._video_writer_class(camera=camera, plugin=self, output_file=video_file)

        frame_queue = Queue()
        frame_processor = threading.Thread(target=self.frame_processor,
                                           kwargs=dict(frame_queue=frame_queue, camera=camera, image_file=image_file))
        frame_processor.start()
        self.fire_event(CameraRecordingStartedEvent(**evt_args))

        try:
            while camera.start_event.is_set():
                if (duration and time.time() - recording_started_time >= duration) \
                        or (n_frames and captured_frames >= n_frames):
                    break

                frame_capture_start = time.time()
                try:
                    frame = self.capture_frame(camera, **kwargs)
                    if not frame:
                        self.logger.warning('Invalid frame received, terminating the capture session')
                        break

                    frame_queue.put(frame)
                except AssertionError as e:
                    self.logger.warning(str(e))
                    continue

                if not n_frames or not camera.info.warmup_seconds or \
                        (time.time() - recording_started_time >= camera.info.warmup_seconds):
                    captured_frames += 1

                if camera.info.fps:
                    wait_time = (1. / camera.info.fps) - (time.time() - frame_capture_start)
                    if wait_time > 0:
                        time.sleep(wait_time)
        finally:
            frame_queue.put(None)
            self.stop_preview(camera)
            for output in camera.get_outputs():
                try:
                    output.close()
                except Exception as e:
                    self.logger.warning('Could not close camera output: {}'.format(str(e)))

            self.close_device(camera, wait_capture=False)
            frame_processor.join(timeout=5.0)
            self.fire_event(CameraRecordingStoppedEvent(**evt_args))

        if image_file:
            self.fire_event(CameraPictureTakenEvent(filename=image_file))

        if video_file:
            self.fire_event(CameraVideoRenderedEvent(filename=video_file))

    def start_camera(self, camera: Camera, preview: bool = False, *args, **kwargs):
        """
        Start a camera capture session.

        :param camera: An initialized :class:`platypush.plugins.camera.Camera` object.
        :param preview: Show a preview of the camera frames.
        """
        assert not (camera.capture_thread and camera.capture_thread.is_alive()), \
            'A capture session is already in progress'

        camera.capture_thread = threading.Thread(target=self.capturing_thread, args=(camera, *args),
                                                 kwargs={'preview': preview, **kwargs})
        camera.capture_thread.start()
        camera.start_event.set()

    @action
    def capture_video(self, duration: Optional[float] = None, video_file: Optional[str] = None, preview: bool = False,
                      **camera) -> Union[str, dict]:
        """
        Capture a video.

        :param duration: Record duration in seconds (default: None, record until ``stop_capture``).
        :param video_file: If set, the stream will be recorded to the specified video file (default: None).
        :param camera: Camera parameters override - see constructors parameters.
        :param preview: Show a preview of the camera frames.
        :return: If duration is specified, the method will wait until the recording is done and return the local path
            to the recorded resource. Otherwise, it will return the status of the camera device after starting it.
        """
        camera = self.open_device(**camera)
        self.start_camera(camera, duration=duration, video_file=video_file, frames_dir=None, image_file=None,
                          preview=preview)

        if duration:
            self.wait_capture(camera)
            return video_file

        return self.status(camera.info.device)

    @action
    def stop_capture(self, device: Optional[Union[int, str]] = None):
        """
        Stop any capturing session on the specified device.

        :param device: Name/path/ID of the device to stop (default: all the active devices).
        """
        devices = self._devices.copy()
        stop_devices = list(devices.values())[:]
        if device:
            stop_devices = [self._devices[device]] if device in self._devices else []

        for device in stop_devices:
            self.close_device(device)

    @action
    def capture_image(self, image_file: str, preview: bool = False, **camera) -> str:
        """
        Capture an image.

        :param image_file: Path where the output image will be stored.
        :param camera: Camera parameters override - see constructors parameters.
        :param preview: Show a preview of the camera frames.
        :return: The local path to the saved image.
        """

        with self.open(**camera) as camera:
            warmup_frames = camera.info.warmup_frames if camera.info.warmup_frames else 1
            self.start_camera(camera, image_file=image_file, n_frames=warmup_frames, preview=preview)
            self.wait_capture(camera)

        return image_file

    # noinspection PyUnusedLocal
    @action
    def take_picture(self, image_file: str, preview: bool = False, **camera) -> str:
        """
        Alias for :meth:`.capture_image`.

        :param image_file: Path where the output image will be stored.
        :param camera: Camera parameters override - see constructors parameters.
        :param preview: Show a preview of the camera frames.
        :return: The local path to the saved image.
        """
        return self.capture_image(image_file, **camera)

    @action
    def capture_sequence(self, duration: Optional[float] = None, n_frames: Optional[int] = None, preview: bool = False,
                         **camera) -> str:
        """
        Capture a sequence of frames from a camera and store them to a directory.

        :param duration: Duration of the sequence in seconds (default: until :meth:`.stop_capture` is called).
        :param n_frames: Number of images to be captured (default: until :meth:`.stop_capture` is called).
        :param camera: Camera parameters override - see constructors parameters. ``frames_dir`` and ``fps`` in
            particular can be specifically tuned for ``capture_sequence``.
        :param preview: Show a preview of the camera frames.
        :return: The directory where the image files have been stored.
        """
        with self.open(**camera) as camera:
            self.start_camera(camera, duration=duration, n_frames=n_frames, preview=preview)
            self.wait_capture(camera)
            return camera.info.frames_dir

    @action
    def capture_preview(self, duration: Optional[float] = None, n_frames: Optional[int] = None, **camera) -> dict:
        """
        Start a camera preview session.

        :param duration: Preview duration (default: until :meth:`.stop_capture` is called).
        :param n_frames: Number of frames to display before closing (default: until :meth:`.stop_capture` is called).
        :param camera: Camera object properties.
        :return: The status of the device.
        """
        camera = self.open_device(frames_dir=None, **camera)
        self.start_camera(camera, duration=duration, n_frames=n_frames, preview=True)
        return self.status(camera.info.device)

    @staticmethod
    def _prepare_server_socket(camera: Camera) -> socket.socket:
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server_socket.bind((  # lgtm [py/bind-socket-all-network-interfaces]
            camera.info.bind_address or '0.0.0.0',
            camera.info.listen_port
        ))
        server_socket.listen(1)
        server_socket.settimeout(1)
        return server_socket

    def _accept_client(self, server_socket: socket.socket) -> Optional[IO]:
        try:
            sock = server_socket.accept()[0]
            self.logger.info('Accepted client connection from {}'.format(sock.getpeername()))
            return sock.makefile('wb')
        except socket.timeout:
            return

    def streaming_thread(self, camera: Camera, stream_format: str, duration: Optional[float] = None):
        streaming_started_time = time.time()
        server_socket = self._prepare_server_socket(camera)
        sock = None
        self.logger.info('Starting streaming on port {}'.format(camera.info.listen_port))

        try:
            while camera.stream_event.is_set():
                if duration and time.time() - streaming_started_time >= duration:
                    break

                sock = self._accept_client(server_socket)
                if not sock:
                    continue

                if camera.info.device not in self._devices:
                    info = camera.info.to_dict()
                    info['stream_format'] = stream_format
                    camera = self.open_device(stream=True, **info)

                camera.stream.sock = sock
                self.start_camera(camera, duration=duration, frames_dir=None, image_file=None)
        finally:
            self._cleanup_stream(camera, server_socket, sock)
            self.logger.info('Stopped camera stream')

    def _cleanup_stream(self, camera: Camera, server_socket: socket.socket, client: IO):
        if client:
            try:
                client.close()
            except Exception as e:
                self.logger.warning('Error on client socket close: {}'.format(str(e)))

        try:
            server_socket.close()
        except Exception as e:
            self.logger.warning('Error on server socket close: {}'.format(str(e)))

        if camera.stream:
            try:
                camera.stream.close()
            except Exception as e:
                self.logger.warning('Error while closing the encoding stream: {}'.format(str(e)))

    @action
    def start_streaming(self, duration: Optional[float] = None, stream_format: str = 'mkv', **camera) -> dict:
        """
        Expose the video stream of a camera over a TCP connection.

        :param duration: Streaming thread duration (default: until :meth:`.stop_streaming` is called).
        :param stream_format: Format of the output stream - e.g. ``h264``, ``mjpeg``, ``mkv`` etc. (default: ``mkv``).
        :param camera: Camera object properties - see constructor parameters.
        :return: The status of the device.
        """
        camera = self.open_device(stream=True, stream_format=stream_format, **camera)
        return self._start_streaming(camera, duration, stream_format)

    def _start_streaming(self, camera: Camera, duration: Optional[float], stream_format: str):
        assert camera.info.listen_port, 'No listen_port specified/configured'
        assert not camera.stream_event.is_set() and camera.info.device not in self._streams, \
            'A streaming session is already running for device {}'.format(camera.info.device)

        self._streams[camera.info.device] = camera
        camera.stream_event.set()

        camera.stream_thread = threading.Thread(target=self.streaming_thread, kwargs=dict(
            camera=camera, duration=duration, stream_format=stream_format))
        camera.stream_thread.start()
        return self.status(camera.info.device)

    @action
    def stop_streaming(self, device: Optional[Union[int, str]] = None):
        """
        Stop a camera over TCP session.

        :param device: Name/path/ID of the device to stop (default: all the active devices).
        """
        streams = self._streams.copy()
        stop_devices = list(streams.values())[:]
        if device:
            stop_devices = [self._streams[device]] if device in self._streams else []

        for device in stop_devices:
            self._stop_streaming(device)

    def _stop_streaming(self, camera: Camera):
        camera.stream_event.clear()
        if camera.stream_thread and camera.stream_thread.is_alive():
            camera.stream_thread.join(timeout=5.0)

        if camera.info.device in self._streams:
            del self._streams[camera.info.device]

    def _status(self, device: Union[int, str]) -> dict:
        camera = self._devices.get(device, self._streams.get(device))
        if not camera:
            return {}

        return {
            **camera.info.to_dict(),
            'active': True if camera.capture_thread and camera.capture_thread.is_alive() else False,
            'capturing': True if camera.capture_thread and camera.capture_thread.is_alive() and
            camera.start_event.is_set() else False,
            'streaming': camera.stream_thread and camera.stream_thread.is_alive() and camera.stream_event.is_set(),
        }

    @action
    def status(self, device: Optional[Union[int, str]] = None):
        """
        Returns the status of the specified camera or all the active cameras if ``device`` is ``None``.
        """

        if device:
            return self._status(device)

        return {
            id: self._status(device)
            for id, camera in self._devices.items()
        }

    @staticmethod
    def transform_frame(frame, color_transform):
        """
        Frame color space (e.g. ``RGB24``, ``YUV`` etc.) transform logic. Does nothing unless implemented by a
        derived plugin.
        """
        return frame.convert(color_transform)

    def to_grayscale(self, frame):
        """
        Convert a frame to grayscale. The default implementation assumes that frame is a ``PIL.Image`` object.

        :param frame: Image frame (default: a ``PIL.Image`` object).
        """
        from PIL import ImageOps
        return ImageOps.grayscale(frame)

    @staticmethod
    def rotate_frame(frame, rotation: Optional[Union[float, int]] = None):
        """
        Frame rotation logic. The default implementation assumes that frame is a ``PIL.Image`` object.

        :param frame: Image frame (default: a ``PIL.Image`` object).
        :param rotation: Rotation angle in degrees.
        """
        if not rotation:
            return frame

        return frame.rotate(rotation, expand=True)

    @staticmethod
    def flip_frame(frame, horizontal_flip: bool = False, vertical_flip: bool = False):
        """
        Frame flip logic. Does nothing unless implemented by a derived plugin.

        :param frame: Image frame (default: a ``PIL.Image`` object).
        :param horizontal_flip: Flip along the horizontal axis.
        :param vertical_flip: Flip along the vertical axis.
        """
        from PIL import Image

        if horizontal_flip:
            frame = frame.transpose(Image.FLIP_TOP_BOTTOM)
        if vertical_flip:
            frame = frame.transpose(Image.FLIP_LEFT_RIGHT)

        return frame

    @staticmethod
    def scale_frame(frame, scale_x: Optional[float] = None, scale_y: Optional[float] = None):
        """
        Frame scaling logic. The default implementation assumes that frame is a ``PIL.Image`` object.

        :param frame: Image frame (default: a ``PIL.Image`` object).
        :param scale_x: X-scale factor.
        :param scale_y: Y-scale factor.
        """
        from PIL import Image
        if not (scale_x and scale_y) or (scale_x == 1 and scale_y == 1):
            return frame

        size = (int(frame.size[0] * scale_x), int(frame.size[1] * scale_y))
        return frame.resize(size, Image.ANTIALIAS)

    @staticmethod
    def encode_frame(frame, encoding: str = 'jpeg') -> bytes:
        """
        Encode a frame to a target type. The default implementation assumes that frame is a ``PIL.Image`` object.

        :param frame: Image frame (default: a ``PIL.Image`` object).
        :param encoding: Image encoding (e.g. ``jpeg``).
        """
        if not encoding:
            return frame

        with io.BytesIO() as buf:
            frame.save(buf, format=encoding)
            return buf.getvalue()

    @staticmethod
    def _get_warmup_seconds(camera: Camera) -> float:
        if camera.info.warmup_seconds:
            return camera.info.warmup_seconds
        if camera.info.warmup_frames and camera.info.fps:
            return camera.info.warmup_frames / camera.info.fps
        return 0


# vim:sw=4:ts=4:et:



import json
from typing import Optional

from redis import Redis

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message import Message


class RedisBackend(Backend):
    """
    Backend that reads messages from a configured Redis queue (default:
    ``platypush_bus_mq``) and posts them to the application bus.  Very
    useful when you have plugin whose code is executed in another process
    and can't post events or requests to the application bus.
    """

    def __init__(self, queue='platypush_bus_mq', redis_args=None, *args, **kwargs):
        """
        :param queue: Queue name to listen on (default: ``platypush_bus_mq``)
        :type queue: str

        :param redis_args: Arguments that will be passed to the redis-py constructor (e.g. host, port, password), see
            https://redis-py.readthedocs.io/en/latest/
        :type redis_args: dict
        """

        super().__init__(*args, **kwargs)
        if redis_args is None:
            redis_args = {}

        self.queue = queue

        if not redis_args:
            redis_plugin = get_plugin('redis')
            if redis_plugin and redis_plugin.kwargs:
                redis_args = redis_plugin.kwargs

        self.redis_args = redis_args
        self.redis: Optional[Redis] = None

    def send_message(self, msg, queue_name=None, **kwargs):
        msg = str(msg)
        if queue_name:
            self.redis.rpush(queue_name, msg)
        else:
            self.redis.rpush(self.queue, msg)

    def get_message(self, queue_name=None):
        queue = queue_name or self.queue
        data = self.redis.blpop(queue, timeout=1)
        if data is None:
            return

        msg = data[1].decode()
        try:
            msg = Message.build(msg)
        except Exception as e:
            self.logger.debug(str(e))
            try:
                import ast
                msg = Message.build(ast.literal_eval(msg))
            except Exception as ee:
                self.logger.debug(str(ee))
                try:
                    msg = json.loads(msg)
                except Exception as eee:
                    self.logger.exception(eee)

        return msg

    def run(self):
        super().run()
        self.logger.info('Initialized Redis backend on queue {} with arguments {}'.format(self.queue, self.redis_args))

        with Redis(**self.redis_args) as self.redis:
            while not self.should_stop():
                try:
                    msg = self.get_message()
                    if not msg:
                        continue

                    self.logger.info('Received message on the Redis backend: {}'.format(msg))
                    self.on_message(msg)
                except Exception as e:
                    self.logger.exception(e)

        self.logger.info('Redis backend terminated')


# vim:sw=4:ts=4:et:



import asyncio
import importlib
import logging

from threading import RLock
from typing import Optional, Any

from ..bus import Bus
from ..config import Config
from ..utils import get_enabled_plugins

logger = logging.getLogger('platypush:context')

# Map: backend_name -> backend_instance
backends = {}

# Map: plugin_name -> plugin_instance
plugins = {}

# Map: plugin_name -> init_lock to make sure that a plugin isn't initialized
# multiple times
plugins_init_locks = {}

# Reference to the main application bus
main_bus = None


def register_backends(bus=None, global_scope=False, **kwargs):
    """Initialize the backend objects based on the configuration and returns
        a name -> backend_instance map.
    Params:
        bus -- If specific (it usually should), the messages processed by the
            backends will be posted on this bus.

        kwargs -- Any additional key-value parameters required to initialize the backends
    """

    global main_bus
    if bus:
        main_bus = bus

    if global_scope:
        global backends
    else:
        backends = {}

    for (name, cfg) in Config.get_backends().items():
        module = importlib.import_module('platypush.backend.' + name)

        # e.g. backend.pushbullet main class: PushbulletBackend
        cls_name = ''
        for token in module.__name__.title().split('.')[2:]:
            cls_name += token.title()
        cls_name += 'Backend'

        try:
            b = getattr(module, cls_name)(bus=bus, **cfg, **kwargs)
            backends[name] = b
        except AttributeError as e:
            logger.warning('No such class in {}: {}'.format(module.__name__, cls_name))
            raise RuntimeError(e)

    return backends


def register_plugins(bus=None):
    from ..plugins import RunnablePlugin

    for plugin in get_enabled_plugins().values():
        if isinstance(plugin, RunnablePlugin):
            plugin.bus = bus
            plugin.start()


def get_backend(name):
    """Returns the backend instance identified by name if it exists"""

    global backends
    return backends.get(name)


def get_plugin(plugin_name, reload=False):
    """Registers a plugin instance by name if not registered already, or
    returns the registered plugin instance"""
    global plugins
    global plugins_init_locks

    if plugin_name not in plugins_init_locks:
        plugins_init_locks[plugin_name] = RLock()

    if plugin_name in plugins and not reload:
        return plugins[plugin_name]

    try:
        plugin = importlib.import_module('platypush.plugins.' + plugin_name)
    except ImportError as e:
        logger.warning('No such plugin: {}'.format(plugin_name))
        raise RuntimeError(e)

    # e.g. plugins.music.mpd main class: MusicMpdPlugin
    cls_name = ''
    for token in plugin_name.split('.'):
        cls_name += token.title()
    cls_name += 'Plugin'

    plugin_conf = (
        Config.get_plugins()[plugin_name] if plugin_name in Config.get_plugins() else {}
    )

    if 'disabled' in plugin_conf:
        if plugin_conf['disabled'] is True:
            return None
        del plugin_conf['disabled']

    if 'enabled' in plugin_conf:
        if plugin_conf['enabled'] is False:
            return None
        del plugin_conf['enabled']

    try:
        plugin_class = getattr(plugin, cls_name)
    except AttributeError as e:
        logger.warning(
            'No such class in {}: {} [error: {}]'.format(plugin_name, cls_name, str(e))
        )
        raise RuntimeError(e)

    with plugins_init_locks[plugin_name]:
        if plugins.get(plugin_name) and not reload:
            return plugins[plugin_name]
        plugins[plugin_name] = plugin_class(**plugin_conf)

    return plugins[plugin_name]


def get_bus() -> Bus:
    global main_bus
    if main_bus:
        return main_bus

    from platypush.bus.redis import RedisBus

    return RedisBus()


def get_or_create_event_loop():
    try:
        loop = asyncio.get_event_loop()
    except (DeprecationWarning, RuntimeError):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    return loop


class Variable:
    """
    Utility class to wrap platform variables in your custom scripts.
    Usage:

        .. code-block:: python

            # Pass `persisted=False` to get/set an in-memory variable
            # on the Redis instance (default: the variable is
            # persisted on the internal database)
            var = Variable('myvar')
            value = var.get()
            var.set('new value')

    """

    def __init__(self, name: str, persisted: bool = True):
        self.name = name
        self._persisted = persisted

    def get(self) -> Optional[Any]:
        plugin = get_plugin('variable')
        getter = getattr(plugin, 'get' if self._persisted else 'mget')
        return getter(self.name).output.get(self.name)

    def set(self, value: Any):
        plugin = get_plugin('variable')
        setter = getattr(plugin, 'set' if self._persisted else 'mset')
        setter(**{self.name: value})


# vim:sw=4:ts=4:et:



import json
from typing import Optional

from platypush.backend.mqtt import MqttBackend
from platypush.context import get_plugin
from platypush.message.event.zigbee.mqtt import ZigbeeMqttOnlineEvent, ZigbeeMqttOfflineEvent, \
        ZigbeeMqttDevicePropertySetEvent, ZigbeeMqttDevicePairingEvent, ZigbeeMqttDeviceConnectedEvent, \
        ZigbeeMqttDeviceBannedEvent, ZigbeeMqttDeviceRemovedEvent, ZigbeeMqttDeviceRemovedFailedEvent, \
        ZigbeeMqttDeviceWhitelistedEvent, ZigbeeMqttDeviceRenamedEvent, ZigbeeMqttDeviceBindEvent, \
        ZigbeeMqttDeviceUnbindEvent, ZigbeeMqttGroupAddedEvent, ZigbeeMqttGroupAddedFailedEvent, \
        ZigbeeMqttGroupRemovedEvent, ZigbeeMqttGroupRemovedFailedEvent, ZigbeeMqttGroupRemoveAllEvent, \
        ZigbeeMqttGroupRemoveAllFailedEvent, ZigbeeMqttErrorEvent


class ZigbeeMqttBackend(MqttBackend):
    """
    Listen for events on a zigbee2mqtt service.

    Triggers:

        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttOnlineEvent` when the service comes online.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttOfflineEvent` when the service goes offline.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDevicePropertySetEvent` when the properties of a
          connected device change.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDevicePairingEvent` when a device is pairing.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceConnectedEvent` when a device connects
          to the network.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceBannedEvent` when a device is banned
          from the network.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceRemovedEvent` when a device is removed
          from the network.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceRemovedFailedEvent` when a request to
          remove a device from the network fails.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceWhitelistedEvent` when a device is
          whitelisted on the network.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceRenamedEvent` when a device is
          renamed on the network.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceBindEvent` when a device bind event
          occurs.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttDeviceUnbindEvent` when a device unbind event
          occurs.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttGroupAddedEvent` when a group is added.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttGroupAddedFailedEvent` when a request to
          add a new group fails.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttGroupRemovedEvent` when a group is removed.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttGroupRemovedFailedEvent` when a request to
          remove a group fails.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttGroupRemoveAllEvent` when all the devices
          are removed from a group.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttGroupRemoveAllFailedEvent` when a request to
          remove all the devices from a group fails.
        * :class:`platypush.message.event.zigbee.mqtt.ZigbeeMqttErrorEvent` when an internal error occurs
          on the zigbee2mqtt service.

    Requires:

        * **paho-mqtt** (``pip install paho-mqtt``)
        * The :class:`platypush.plugins.zigbee.mqtt.ZigbeeMqttPlugin` plugin configured.

    """

    def __init__(self, host: Optional[str] = None, port: Optional[int] = None, base_topic='zigbee2mqtt',
                 tls_cafile: Optional[str] = None, tls_certfile: Optional[str] = None,
                 tls_keyfile: Optional[str] = None, tls_version: Optional[str] = None,
                 tls_ciphers: Optional[str] = None, username: Optional[str] = None,
                 password: Optional[str] = None, client_id: Optional[str] = None, *args, **kwargs):
        """
        :param host: MQTT broker host (default: host configured on the ``zigbee.mqtt`` plugin).
        :param port: MQTT broker port (default: 1883).
        :param base_topic: Prefix of the topics published by zigbee2mqtt (default: '``zigbee2mqtt``').
        :param tls_cafile: If TLS/SSL is enabled on the MQTT server and the certificate requires a certificate authority
            to authenticate it, `ssl_cafile` will point to the provided ca.crt file (default: None)
        :param tls_certfile: If TLS/SSL is enabled on the MQTT server and a client certificate it required, specify it
            here (default: None)
        :param tls_keyfile: If TLS/SSL is enabled on the MQTT server and a client certificate key it required,
            specify it here (default: None) :type tls_keyfile: str
        :param tls_version: If TLS/SSL is enabled on the MQTT server and it requires a certain TLS version, specify it
            here (default: None)
        :param tls_ciphers: If TLS/SSL is enabled on the MQTT server and an explicit list of supported ciphers is
            required, specify it here (default: None)
        :param username: Specify it if the MQTT server requires authentication (default: None)
        :param password: Specify it if the MQTT server requires authentication (default: None)
        :param client_id: MQTT client ID (default: ``<device_id>-zigbee-mqtt``, to prevent clashes with the
            :class:`platypush.backend.mqtt.MqttBackend` ``client_id``.
        """

        plugin = get_plugin('zigbee.mqtt')
        self.base_topic = base_topic or plugin.base_topic
        self._devices = {}
        self._groups = {}
        self._last_state = None
        self.server_info = {
            'host': host or plugin.host,
            'port': port or plugin.port or self._default_mqtt_port,
            'tls_cafile': tls_cafile or plugin.tls_cafile,
            'tls_certfile': tls_certfile or plugin.tls_certfile,
            'tls_ciphers': tls_ciphers or plugin.tls_ciphers,
            'tls_keyfile': tls_keyfile or plugin.tls_keyfile,
            'tls_version': tls_version or plugin.tls_version,
            'username': username or plugin.username,
            'password': password or plugin.password,
        }

        kwargs = {
            **kwargs,
            **self.server_info,
        }

        listeners = [{
            **self.server_info,
            'topics': [
                self.base_topic + '/' + topic
                for topic in ['bridge/state', 'bridge/log', 'bridge/logging', 'bridge/devices', 'bridge/groups']
            ],
        }]

        super().__init__(
            *args, subscribe_default_topic=False,
            listeners=listeners, client_id=client_id, **kwargs
        )

        if not client_id:
            self.client_id += '-zigbee-mqtt'

    def _process_state_message(self, client, msg):
        if msg == self._last_state:
            return

        if msg == 'online':
            evt = ZigbeeMqttOnlineEvent
        elif msg == 'offline':
            evt = ZigbeeMqttOfflineEvent
            self.logger.warning('zigbee2mqtt service is offline')
        else:
            return

        # noinspection PyProtectedMember
        self.bus.post(evt(host=client._host, port=client._port))
        self._last_state = msg

    def _process_log_message(self, client, msg):
        msg_type = msg.get('type')
        text = msg.get('message')
        # noinspection PyProtectedMember
        args = {'host': client._host, 'port': client._port}

        if msg_type == 'devices':
            devices = {}
            for dev in (text or []):
                devices[dev['friendly_name']] = dev
                client.subscribe(self.base_topic + '/' + dev['friendly_name'])
        elif msg_type == 'pairing':
            self.bus.post(ZigbeeMqttDevicePairingEvent(device=text, **args))
        elif msg_type in ['device_ban', 'device_banned']:
            self.bus.post(ZigbeeMqttDeviceBannedEvent(device=text, **args))
        elif msg_type in ['device_removed_failed', 'device_force_removed_failed']:
            force = msg_type == 'device_force_removed_failed'
            self.bus.post(ZigbeeMqttDeviceRemovedFailedEvent(device=text, force=force, **args))
        elif msg_type == 'device_whitelisted':
            self.bus.post(ZigbeeMqttDeviceWhitelistedEvent(device=text, **args))
        elif msg_type == 'device_renamed':
            self.bus.post(ZigbeeMqttDeviceRenamedEvent(device=text, **args))
        elif msg_type == 'device_bind':
            self.bus.post(ZigbeeMqttDeviceBindEvent(device=text, **args))
        elif msg_type == 'device_unbind':
            self.bus.post(ZigbeeMqttDeviceUnbindEvent(device=text, **args))
        elif msg_type == 'device_group_add':
            self.bus.post(ZigbeeMqttGroupAddedEvent(group=text, **args))
        elif msg_type == 'device_group_add_failed':
            self.bus.post(ZigbeeMqttGroupAddedFailedEvent(group=text, **args))
        elif msg_type == 'device_group_remove':
            self.bus.post(ZigbeeMqttGroupRemovedEvent(group=text, **args))
        elif msg_type == 'device_group_remove_failed':
            self.bus.post(ZigbeeMqttGroupRemovedFailedEvent(group=text, **args))
        elif msg_type == 'device_group_remove_all':
            self.bus.post(ZigbeeMqttGroupRemoveAllEvent(group=text, **args))
        elif msg_type == 'device_group_remove_all_failed':
            self.bus.post(ZigbeeMqttGroupRemoveAllFailedEvent(group=text, **args))
        elif msg_type == 'zigbee_publish_error':
            self.logger.error('zigbee2mqtt error: {}'.format(text))
            self.bus.post(ZigbeeMqttErrorEvent(error=text, **args))
        elif msg.get('level') in ['warning', 'error']:
            log = getattr(self.logger, msg['level'])
            log('zigbee2mqtt {}: {}'.format(msg['level'], text or msg.get('error', msg.get('warning'))))

    def _process_devices(self, client, msg):
        devices_info = {
            device.get('friendly_name', device.get('ieee_address')): device
            for device in msg
        }

        # noinspection PyProtectedMember
        event_args = {'host': client._host, 'port': client._port}
        client.subscribe(*[
            self.base_topic + '/' + device
            for device in devices_info.keys()
        ])

        for name, device in devices_info.items():
            if name not in self._devices:
                self.bus.post(ZigbeeMqttDeviceConnectedEvent(device=name, **event_args))

            exposes = (device.get('definition', {}) or {}).get('exposes', [])
            client.publish(
                self.base_topic + '/' + name + '/get',
                json.dumps(get_plugin('zigbee.mqtt').build_device_get_request(exposes))
            )

        devices_copy = [*self._devices.keys()]
        for name in devices_copy:
            if name not in devices_info:
                self.bus.post(ZigbeeMqttDeviceRemovedEvent(device=name, **event_args))
                del self._devices[name]

        self._devices = {device: {} for device in devices_info.keys()}

    def _process_groups(self, client, msg):
        # noinspection PyProtectedMember
        event_args = {'host': client._host, 'port': client._port}
        groups_info = {
            group.get('friendly_name', group.get('id')): group
            for group in msg
        }

        for name in groups_info.keys():
            if name not in self._groups:
                self.bus.post(ZigbeeMqttGroupAddedEvent(group=name, **event_args))

        groups_copy = [*self._groups.keys()]
        for name in groups_copy:
            if name not in groups_info:
                self.bus.post(ZigbeeMqttGroupRemovedEvent(group=name, **event_args))
                del self._groups[name]

        self._groups = {group: {} for group in groups_info.keys()}

    def on_mqtt_message(self):
        def handler(client, _, msg):
            topic = msg.topic[len(self.base_topic)+1:]
            data = msg.payload.decode()
            if not data:
                return

            try:
                data = json.loads(data)
            except (ValueError, TypeError):
                pass

            if topic == 'bridge/state':
                self._process_state_message(client, data)
            elif topic in ['bridge/log', 'bridge/logging']:
                self._process_log_message(client, data)
            elif topic == 'bridge/devices':
                self._process_devices(client, data)
            elif topic == 'bridge/groups':
                self._process_groups(client, data)
            else:
                suffix = topic.split('/')[-1]
                if suffix not in self._devices:
                    return

                name = suffix
                changed_props = {k: v for k, v in data.items() if v != self._devices[name].get(k)}

                if changed_props:
                    # noinspection PyProtectedMember
                    self.bus.post(ZigbeeMqttDevicePropertySetEvent(host=client._host, port=client._port,
                                                                   device=name, properties=changed_props))

                self._devices[name].update(data)

        return handler

    def run(self):
        super().run()


# vim:sw=4:ts=4:et:



from typing import Optional

from platypush.message.event import Event


class NewWeatherConditionEvent(Event):
    """
    Event triggered when the weather condition changes
    """

    def __init__(self, *args, plugin_name: Optional[str] = None, **kwargs):
        super().__init__(*args, plugin_name=plugin_name, **kwargs)


class NewPrecipitationForecastEvent(Event):
    """
    Event triggered when the precipitation forecast changes
    """
    def __init__(self, *args, plugin_name: Optional[str] = None, average: float, total: float,
                 time_frame: int, **kwargs):
        super().__init__(*args, plugin_name=plugin_name, average=average, total=total, time_frame=time_frame, **kwargs)


# vim:sw=4:ts=4:et:



import os
import time
from typing import Optional, Tuple, Union, Dict, Callable, Any

from samsungtvws import SamsungTVWS

from platypush.config import Config
from platypush.plugins import Plugin, action


class TvSamsungWsPlugin(Plugin):
    """
    Control a Samsung smart TV with Tizen OS over WiFi/ethernet. It should support any post-2016 Samsung with Tizen OS
    and enabled websocket-based connection.

    Requires:

        * **samsungtvws** (``pip install samsungtvws``)

    """

    def __init__(self, host: Optional[str] = None, port: int = 8002, timeout: Optional[int] = 5, name='platypush',
                 token_file: Optional[str] = None, **kwargs):
        """
        :param host: IP address or host name of the smart TV.
        :param port: Websocket port (default: 8002).
        :param timeout: Connection timeout in seconds (default: 5, specify 0 or None for no timeout).
        :param name: Name of the remote device (default: platypush).
        :param token_file: Path to the token file (default: ``~/.local/share/platypush/samsungtvws/token.txt``)
        """
        super().__init__(**kwargs)
        self.workdir = os.path.join(Config.get('workdir'), 'samsungtvws')
        if not token_file:
            token_file = os.path.join(self.workdir, 'token.txt')

        self.host = host
        self.port = port
        self.timeout = timeout
        self.name = name
        self.token_file = token_file
        self._connections: Dict[Tuple[host, port], SamsungTVWS] = {}
        os.makedirs(self.workdir, mode=0o700, exist_ok=True)

    def _get_host_and_port(self, host: Optional[str] = None, port: Optional[int] = None) -> Tuple[str, int]:
        host = host or self.host
        port = port or self.port
        assert host and port, 'No host/port specified'
        return host, port

    def connect(self, host: Optional[str] = None, port: Optional[int] = None) -> SamsungTVWS:
        host, port = self._get_host_and_port(host, port)
        if (host, port) not in self._connections:
            self._connections[(host, port)] = SamsungTVWS(host=host, port=port, token_file=self.token_file,
                                                          timeout=self.timeout, name=self.name)

        return self._connections[(host, port)]

    def exec(self, func: Callable[[SamsungTVWS], Any], host: Optional[str] = None, port: Optional[int] = None,
             n_tries=2) -> Any:
        tv = self.connect(host, port)

        try:
            return func(tv)
        except Exception as e:
            self.logger.warning(str(e))
            if n_tries <= 0:
                raise e
            else:
                time.sleep(1)
                return self.exec(func, host, port, n_tries-1)

    @action
    def power(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send power on/off control to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().power(), host=host, port=port)

    @action
    def volume_up(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send volume up control to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().volume_up(), host=host, port=port)

    @action
    def volume_down(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send volume down control to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().volume_down(), host=host, port=port)

    @action
    def back(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send back key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().back(), host=host, port=port)

    @action
    def channel(self, channel: int, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Change to the selected channel.

        :param channel: Channel index.
        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().channel(channel), host=host, port=port)

    @action
    def channel_up(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send channel_up key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().channel_up(), host=host, port=port)

    @action
    def channel_down(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send channel_down key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().channel_down(), host=host, port=port)

    @action
    def enter(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send enter key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().enter(), host=host, port=port)

    @action
    def guide(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send guide key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().guide(), host=host, port=port)

    @action
    def home(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send home key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().home(), host=host, port=port)

    @action
    def info(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send info key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().info(), host=host, port=port)

    @action
    def menu(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send menu key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().menu(), host=host, port=port)

    @action
    def mute(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send mute key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().mute(), host=host, port=port)

    @action
    def source(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send source key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().source(), host=host, port=port)

    @action
    def tools(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send tools key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().tools(), host=host, port=port)

    @action
    def up(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send up key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().up(), host=host, port=port)

    @action
    def down(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send down key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().down(), host=host, port=port)

    @action
    def left(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send left key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().left(), host=host, port=port)

    @action
    def right(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send right key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().right(), host=host, port=port)

    @action
    def blue(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send blue key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().blue(), host=host, port=port)

    @action
    def green(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send green key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().green(), host=host, port=port)

    @action
    def red(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send red key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().red(), host=host, port=port)

    @action
    def yellow(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send red key to the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().yellow(), host=host, port=port)

    @action
    def digit(self, digit: int, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Send a digit key to the device.

        :param digit: Digit to send.
        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.shortcuts().digit(digit), host=host, port=port)

    @action
    def run_app(self, app_id: Union[int, str], host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Run an app by ID.

        :param app_id: App ID.
        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        tv = self.connect(host, port)
        tv.rest_app_run(str(app_id))

    @action
    def close_app(self, app_id: Union[int, str], host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Close an app.

        :param app_id: App ID.
        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        tv = self.connect(host, port)
        tv.rest_app_close(str(app_id))

    @action
    def install_app(self, app_id: Union[int, str], host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Install an app.

        :param app_id: App ID.
        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        tv = self.connect(host, port)
        tv.rest_app_install(str(app_id))

    @action
    def status_app(self, app_id: Union[int, str], host: Optional[str] = None, port: Optional[int] = None) -> dict:
        """
        Get the status of an app.

        :param app_id: App ID.
        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        tv = self.connect(host, port)
        return tv.rest_app_status(str(app_id))

    @action
    def list_apps(self, host: Optional[str] = None, port: Optional[int] = None) -> list:
        """
        Get the list of installed apps.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.app_list(), host=host, port=port)

    @action
    def open_browser(self, url: str, host: Optional[str] = None, port: Optional[int] = None) -> None:
        """
        Open a URL in the browser.

        :param url: URL to open.
        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        return self.exec(lambda tv: tv.open_browser(url), host=host, port=port)

    @action
    def device_info(self, host: Optional[str] = None, port: Optional[int] = None) -> dict:
        """
        Return the info of the device.

        :param host: Default host IP/name override.
        :param port: Default port override.
        """
        tv = self.connect(host, port)
        return tv.rest_device_info()


# vim:sw=4:ts=4:et:



from flask import Blueprint, render_template

from platypush.backend.http.app import template_folder
from platypush.backend.http.app.utils import authenticate
from platypush.backend.http.utils import HttpUtils

index = Blueprint('index', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    index,
]


@index.route('/')
@authenticate()
def index():
    """ Route to the main web panel """
    return render_template('index.html', utils=HttpUtils)


# vim:sw=4:ts=4:et:
